# $Id: __init__.py f321b8c1cbca 2012/02/21 11:32:32 patrick $
"""LePrisme engine."""

import re
from os import walk, listdir, makedirs, remove
from os.path import join, exists, splitext, dirname, basename, isdir, relpath
from os.path import getmtime
from shutil import copy, rmtree
from ConfigParser import ConfigParser
from imp import load_source
from lxml import etree
from zipfile import is_zipfile

from ...utils import _, copy_content, unzip, config_get
from ...xml import XML_NS, load
from .publiset import Publiset
from .transform import Transform


# =============================================================================
class Engine(object):
    """Main class for LePrisme engine."""

    # -------------------------------------------------------------------------
    def __init__(self, build):
        """Constructor method.

        :param build: (:class:`~.lib.build.agent.AgentBuild`)
            Main Build object.
        """
        # Attributes
        self.build = build

        # Configuration
        name = join(build.path, 'Processor', 'leprisme.ini')
        if exists(name):
            self._config = ConfigParser({'here': dirname(name),
                'fid': '{fid}', 'ocffile': '{ocffile}'})
            self._config.read(name)
        else:
            build.stopped(_('File "leprisme.ini" is missing.'))
            return

        # Relax NG, scripts, publiset and transformation
        self.relaxngs = self._load_relaxngs()
        self.scripts = self._load_scripts()
        self._transform = Transform(self)

    # -------------------------------------------------------------------------
    def start(self):
        """Start the engine."""
        if self.build.stopped():
            return

        # Initialization
        subdir = self.build.processing['variables'].get('subdir', '') != ''
        output = join(self.build.path, 'Output')
        if not subdir:
            self.initialize(output)
        if self.build.stopped():
            return

        # Process each file
        files = self._file_list()
        if not len(files):
            self.build.stopped(_('nothing to do!'), 'a_error')
            return
        for count, name in enumerate(files):
            percent = 100 * (count + 1) / (len(files) + 1)
            self._process(percent, name)
            if self.build.stopped():
                break

        # Finalization
        if not subdir:
            self.finalize(output)

    # -------------------------------------------------------------------------
    def config(self, section, option, default=None):
        """Retrieve a value from a configuration object.

        :param section: (string)
            Section name.
        :param option: (string)
            Option name.
        :param default: (string, optional)
            Default value
        :return: (string)
            Read value or default value.
        """
        return config_get(self._config, section, option, default)

    # -------------------------------------------------------------------------
    def initialize(self, output):
        """Initialization.

        :param output: (string)
            Full path to output directory.

        ``self.build.processing['templates']`` and
        ``self.build.pack['templates']`` are lists of tuples such as
        ``(<input_file>, <output_path>)``.
        """
        # Create directories
        if not exists(output):
            makedirs(output)
        self._remove_temporary_files(output)
        names = self.config('Initialization', 'directories', '')
        for name in names.split():
            if not exists(join(output, name)):
                makedirs(join(output, name))

        # Copy template files from INI files
        names = self.config('Initialization', 'templates', '')
        for name in names.split():
            path = self.config('template:%s' % name, 'path', '')
            exclude = self.config('template:%s' % name, 'exclude', '')
            copy_content(
                join(self.build.path, 'Processor', 'Templates', name),
                join(output, path), exclude.split())

        # Copy template files from processing and pack templates
        for name, path in self.build.processing['templates'] \
                + self.build.pack['templates']:
            name = join(self.build.data_path, name)
            do_unzip = path[0:6] == 'unzip:'
            path = do_unzip and join(output, path[6:]) or join(output, path)
            if isdir(name):
                copy_content(name, path)
            elif do_unzip and is_zipfile(name):
                unzip(name, path)
            else:
                if not exists(dirname(path)):
                    makedirs(dirname(path))
                copy(name, path)

        # Run initialization script
        if 'initialization' in self.scripts:
            self.scripts['initialization'](self, output)

    # -------------------------------------------------------------------------
    def finalize(self, output):
        """Finalization.

        :param output: (string)
            Full path to output directory.
        """
        # Remove temporary files
        if not self.build.processing['variables'].get('keeptmp'):
            self._remove_temporary_files(output)

        # Run finalization script
        if 'finalization' in self.scripts:
            self.scripts['finalization'](self, output)

    # -------------------------------------------------------------------------
    def _load_relaxngs(self):
        """Load Relax NG.

        :return: (dictionary)
            A dictionary of :class:`lxml.etree.RelaxNG` objets.
        """
        relaxngs = {}
        if not self._config.has_section('RelaxNG'):
            return relaxngs

        for root, filename in self._config.items('RelaxNG'):
            if root not in ('here', 'fid', 'ocffile'):
                try:
                    relaxngs[root] = etree.RelaxNG(etree.parse(filename))
                except IOError, err:
                    self.build.stopped(err)
                    return
                except (etree.XMLSyntaxError, etree.RelaxNGParseError), err:
                    self.build.stopped('"%s": %s' % (filename, err))
                    return
        return relaxngs

    # -------------------------------------------------------------------------
    def _load_scripts(self):
        """Load script files.

        :return: (dictionary)
            A dictionary of script main functions.
         """
        scripts = {}
        for name in (('Initialization', 'script'),
                     ('Transformation', 'preprocess'),
                     ('Transformation', 'postprocess'),
                     ('Finalization', 'script')):
            # Find file
            filename = self.config(name[0], name[1])
            if not filename:
                continue
            if not exists(filename):
                self.build.stopped('Unknown file "%s"' % filename)
                return scripts

            # Load module
            module = load_source(splitext(basename(filename))[0], filename)
            scripts[name[0].lower() if name[1] == 'script' else name[1]] = \
                module.main
        return scripts

    # -------------------------------------------------------------------------
    def _file_list(self):
        """List files in pack according to settings.

        :return: (list)
            File list.
        """
        regex = re.compile(self._config.get('Input', 'file_regex'))
        input_is_dir = self._config.has_option('Input', 'is_dir') \
                       and self._config.get('Input', 'is_dir') == 'true'
        files = []

        for root in self.build.pack['files']:
            fullname = join(self.build.data_path, root)
            if not exists(fullname):
                self.build.stopped('Unknown file "%s"' % root, 'a_warn')
                continue

            if regex.search(root) and isdir(fullname) == input_is_dir:
                files.append(root)
            if not self.build.pack.get('recursive') or not isdir(fullname):
                continue

            for path, dirs, filenames in walk(fullname):
                names = input_is_dir and dirs or filenames
                for name in names:
                    if regex.search(name) and name != 'schemas.xml':
                        name = relpath(join(path, name), self.build.data_path)
                        name = isinstance(name, str) \
                               and unicode(name.decode('utf8')) or name
                        files.append(name)
        return files

    # -------------------------------------------------------------------------
    def _process(self, percent, filename):
        """Process one XML file.

        :param percent: (integer)
            Percent of progress.
        :param filename: (string)
            Relative path of the file to process.
        """
        if self.build.stopped():
            return

        # Load path
        fid = splitext(basename(filename))[0]
        fullname = join(self.build.data_path, filename)
        self.build.log(u'%s ............' % fid, 'a_build', percent)

        # Load folder
        if isdir(fullname):
            self._transform.start(filename, fid, filename, getmtime(fullname))
            return

        # Load file
        if splitext(filename)[1].lower() == '.xml':
            relaxngs = self.relaxngs \
                if self.config('Input', 'validate') == 'true' else None
            data = load(fullname, relaxngs)
            if isinstance(data, basestring):
                self.build.stopped(data, 'a_error')
                return
        else:
            with open(fullname, 'r') as hdl:
                data = hdl.read()

        # Non Publiset file
        # pylint: disable = I0011, E1103
        if isinstance(data, basestring) or data.getroot().tag != 'publiset':
            self._transform.start(filename, fid, data, getmtime(fullname))
            return

        # Publiset selection
        publiset = Publiset(self, dirname(fullname))
        for set_root in data.findall('selection'):
            for file_elt in set_root.xpath('.//file'):
                name = relpath(publiset.fullname(file_elt),
                               self.build.data_path)
                if name != filename:
                    self._process(percent, name)
                if self.build.stopped():
                    return
            fid = set_root.get('%sid' % XML_NS)
            self.build.log('%s ............' % fid, 'a_build', percent)
            self._transform.start(
                filename, fid, publiset.create(set_root), getmtime(fullname))

        # Publiset composition
        for set_root in data.findall('composition'):
            fid = set_root.get('%sid' % XML_NS)
            self.build.log(_('${f}: document composition', {'f': fid}))
            self._transform.start(
                filename, fid, publiset.compose(filename, set_root))
            if self.build.stopped():
                return

    # -------------------------------------------------------------------------
    def _remove_temporary_files(self, output):
        """Remove temporary files.

        :param output: (string)
            Full path to output directory.
        """
        regex = self.config('Finalization', 'remove_regex', '~\.?\w{0,4}$')
        for path, dirs, files in walk(output, topdown=False):
            for name in dirs:
                if (regex and re.search(regex, name)) \
                       or not listdir(join(path, name)):
                    rmtree(join(path, name))
            for name in files:
                if regex and re.search(regex, name):
                    remove(join(path, name))



# $Id: handler.py 513105aef084 2012/02/13 16:15:59 patrick $
"""A *handler* manages access to a storage.
"""

from os import listdir, walk
from os.path import join, getsize, splitext
from time import time, sleep
from threading import Thread
from locale import strxfrm

from .utils import _, decrypt, unzip, get_mime_type
from .utils import age, size_label, cache_method
from .vcs.none import VcsNone
from .vcs.hg import VcsMercurial, VcsLocal


# =============================================================================
class HandlerManager(object):
    """This class manages all handlers.

    One instance of :class:`HandlerManager` is created during application
    initialization. It is only used in front mode. It is stored in application
    registry.

    ``self.cache_manager`` is a :class:`beaker.cache.CacheManager` instance.

    ``self._handlers`` is a dictionary of :class:`Handler` objects.
    """

    # -------------------------------------------------------------------------
    def __init__(self, settings, cache_manager):
        """Constructor method.

        :param settings: (dictionary)
            Setting dictionary.
        :param cache_manager: (:class:`beaker.cache.CacheManager` instance)
            Global Beaker cache manager.
        """
        # Attributes
        self.settings = settings
        self.cache_manager = cache_manager
        self._handlers = {}

        # VCS engine classes
        vcs_classes = {'none': VcsNone, 'local': VcsLocal, 'hg': VcsMercurial}
        self.vcs_classes = {}
        for vcs in settings.get('storage.available_vcs', '').split():
            if vcs in vcs_classes and not vcs in self.vcs_classes:
                self.vcs_classes[vcs] = vcs_classes[vcs]

    # -------------------------------------------------------------------------
    def vcs_list(self):
        """Return a list of available Version Control System."""
        return self.vcs_classes.keys()

    # -------------------------------------------------------------------------
    def get_handler(self, storage_id, storage=None):
        """Create or retrieve a Storage Control System for storage
        ``storage``.

        :param storage_id: (string)
            Storage ID.
        :param storage: (:class:`~.models.storages.Storage` instance,
            optional).
        :return: (:class:`Handler` instance or ``None``)
        """
        self._cleanup()
        if storage_id in self._handlers:
            self._handlers[storage_id].expire = \
                time() + self._handlers[storage_id].cache.expiretime
            return self._handlers[storage_id]

        if storage is None or storage_id != storage.storage_id \
               or not storage.vcs_engine in self.vcs_classes:
            return

        self._handlers[storage_id] = Handler(self, storage)
        return self._handlers[storage_id]

    # -------------------------------------------------------------------------
    def progress(self, storage_ids, pending=False):
        """Return the progress of actions on storages.

        :param storage_ids: (list)
            Storage ID list.
        :param pending: (boolean, default=False)
            ``True`` if there is a pending work.
        :return: (tuple)
            Return a tuple such as ``(working, progress_list)``.

        ``working`` is a boolean indicating whether one of the storage is on
        progress. ``progress_list`` is a list of items like ``(status,
        message)``. See :class:`Handler` class.
        """
        self._cleanup()
        working = False
        prgrss = {}
        for storage_id in storage_ids:
            if storage_id in self._handlers:
                handler_prgrss = self._handlers[storage_id].progress()
                if handler_prgrss[0] != 'wait':
                    prgrss[storage_id] = handler_prgrss
                    working |= handler_prgrss[0] == 'run'
        return working | pending, prgrss

    # -------------------------------------------------------------------------
    def _cleanup(self):
        """Delete expired Handler."""
        now = time()
        for storage_id in self._handlers.keys():
            if self._handlers[storage_id].expire < now:
                if self._handlers[storage_id].progress()[0] != 'run':
                    del self._handlers[storage_id]


# =============================================================================
class Handler(object):
    """This class manages access to one storage.

    ``self.uid`` is the ID of the associated storage.

    ``self.expire`` is the deadline for this object.

    ``self.cache`` is a :class:`beaker.cache.Cache` instance.

    ``self.vcs`` is a :class:`~.lib.vcs.Vcs` instance.

    ``self._report`` is a tuple such as ``(status, message, expire, ttl)``
    where ``expire`` is the report validity date and ``status`` is one of the
    following:

    ``self._refresh`` is a tuple such as ``(time_to_refresh, refresh_period)``.

    * ``wait``: waiting
    * ``run``: in progress
    * ``error``: ended with error
    * ``end``: ended with success
    """

    # -------------------------------------------------------------------------
    def __init__(self, handler_manager, storage):
        """Constructor method.

        :param handler_manager: (:class:`HandlerManager` instance)
            Application :class:`HandlerManager` object.
        :param storage: (:class:`~.models.storages.Storage` instance).
        """
        self.uid = storage.storage_id
        self.cache = handler_manager.cache_manager.get_cache(
            'stg_%s' % self.uid,
            expire=int(handler_manager.settings.get('storage.cache', 3600)))
        self.expire = time() + self.cache.expiretime
        self.vcs = handler_manager.vcs_classes[storage.vcs_engine](
            join(handler_manager.settings['storage.root'], self.uid),
            storage.vcs_url, storage.vcs_user, decrypt(storage.vcs_password,
            handler_manager.settings['auth.key']))
        self._report = ('wait', None, 0,
            int(handler_manager.settings.get('storage.report_ttl', 12)))
        self._refresh = [0, int(storage.refresh)]
        self._thread = None

   # -------------------------------------------------------------------------
    def clone(self, request=None):
        """Launch an action to clone a storage.

        :param request: (:class:`pyramid.request.Request` instance, optional)
            Current request.
        """
        # Action directly...
        if request is None:
            return self.vcs.clone()

        # ...or in a thread
        if self.launch(request, self.vcs.clone):
            self._refresh[0] = time() + self._refresh[1]

    # -------------------------------------------------------------------------
    def synchronize(self, request, force=False):
        """Launch an action to synchronize storage with its source.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param force: (boolean, optional)
            Force synchronization even if period is not over.
        :return: (boolean)
            ``True`` if it really launch a synchronization.

        If ``force`` is ``False``, the synchronizaton is only done if the
        delay of ``synchro_period`` seconds has expired.
        """
        # Something to do?
        if not force and self._refresh[0] > time():
            self.expire = time() + self.cache.expiretime
            return False

        # Action directly...
        if request is None:
            self.report('run')
            if not self.vcs.pull_update(self):
                self._refresh[0] = time() + self._refresh[1]
                return True
            return False

        # ...or in a thread
        if self.launch(request, self.vcs.pull_update):
            self._refresh[0] = time() + self._refresh[1]
            return True
        return False

    # -------------------------------------------------------------------------
    def report(self, status, message=None):
        """Save a report.

        :param status: (string)
            Current status. See :class:`Handler` class.
        :param message: (string, optional)
            Message to write in report.
        :return: (string or ``None``)
            Message.
        """
        self._report = (
            status, message, time() + self._report[3], self._report[3])
        self.expire = time() + self.cache.expiretime
        return message

    # -------------------------------------------------------------------------
    def progress(self):
        """Return the progress of action on the storage.

        :return: (tuple)
            A tuple such as ``(status, message)``. See :class:`Handler` class.
        """
        if self._thread is not None and self._thread.is_alive():
            return ('run', None)
        if self._report[0] != 'wait' and self._report[2] < time():
            self.report('wait')
        return self._report[0:2]

    # -------------------------------------------------------------------------
    @cache_method()
    def dir_infos(self, path, sort):
        """List all files of a directory with VCS informations.

        :param path: (string)
            Relative path of the directory.
        :param sort: (string)
            A sort order among ``+name``, ``-name``, ``+size``, ``-size``,
            ``+date``, ``-date``.
        :return: (tuple)
            A tuple of list such as ``(dirs, files)``.

        See: :meth:`file_infos`.
        """
        dirs, files = walk(join(self.vcs.path, path)).next()[1:]
        dirs = self.file_infos(path, dirs)
        files = self.file_infos(path, files)

        # Sort
        key = {'size': lambda k: k[2], 'date': lambda k: k[3],
               }.get(sort[1:], lambda k: strxfrm(k[0].encode('utf8')))
        dirs = sorted(dirs, key=key, reverse=(sort[0] == '-'))
        files = sorted(files, key=key, reverse=(sort[0] == '-'))

        # Labels
        for item in dirs:
            item[2] = size_label(item[2], True)
            item[3] = age(item[3])
        for item in files:
            item[2] = size_label(item[2], False)
            item[3] = age(item[3])

        return dirs, files

    # -------------------------------------------------------------------------
    def file_infos(self, path, filenames):
        """Return a sorted list of informations such as
        ``(file_name, file_type, size, revision, user, date, message)`` on each
        file in ``filenames``.

        :param path: (string)
            Relative path to files.
        :param filenames: (list)
            List of files.
        :return: (list of tuples)
            A list of lists such as ``['README', 'plain', '45.7 Kio',
            '2 days', '16:0e0229a916f4', 'user1', 'Bug fixed']``
        """
        file_infos = []
        for filename in filenames:
            vcs_infos = self.vcs.file_infos(path, filename)
            if vcs_infos is None:
                continue
            fullname = join(self.vcs.path, path, filename)
            filetype = get_mime_type(fullname)[1]
            size = len(listdir(fullname)) if filetype == 'folder' \
                   else getsize(fullname)
            file_infos.append([filename, filetype, size,
                vcs_infos[0], vcs_infos[1], vcs_infos[2], vcs_infos[3]])

        return file_infos

    # -------------------------------------------------------------------------
    def upload(self, user, path, upload_file, filename, message):
        """Synchronize, remove files and propagate.

        :param user: (list)
            VCS user like ``(user_id, password, user_name)``.
        :param path: (string)
            Relative path to files.
        :param filename: (:class:`cgi.FieldStorage` instance)
            FieldStorage of the file to upload.
        :param filename: (string or `None``)
            Name of file to replace or ``None``.
        :param message: (string)
            Commit message.
        """
        # Check filename
        if filename is not None and filename != upload_file.filename:
            self.report('error', _('File names are different.'))
            return

        # Check repository
        if not self._check_repository(user, message):
            return

        # Upload files
        fullpath = join(self.vcs.path, path or '.')
        ext = splitext(upload_file.filename)[1]
        if ext != '.zip' or filename:
            filename = filename or upload_file.filename
            with open(join(fullpath, filename), 'w') as hdl:
                hdl.write(upload_file.file.read())
        else:
            unzip(upload_file.file, fullpath)

        # Add
        if self.vcs.add(path, self):
            return

        # Propagate
        if not self.vcs.commit_push(message, user[0], user[1], user[2], self):
            self.report('end')

    # -------------------------------------------------------------------------
    def mkdir(self, path, name):
        """Make a directory.

        :param path: (string)
            Relative path to directory to create.
        :param name: (string)
            Name of directory to create.
        """
        if not self.synchronize(None, True):
            return
        if not self.vcs.mkdir(path, name, self):
            self.report('end')

    # -------------------------------------------------------------------------
    def add(self, user, path, message):
        """Synchronize, remove files and propagate.

        :param user: (list)
            VCS user like ``(user_id, password, user_name)``.
        :param path: (string)
            Relative path to files.
        :param message: (string)
            Commit message.
        """
        # Check repository
        if not self._check_repository(user, message):
            return

        # Add
        if self.vcs.add(path, self):
            return

        # Propagate
        if not self.vcs.commit_push(message, user[0], user[1], user[2], self):
            self.report('end')

    # -------------------------------------------------------------------------
    def rename(self, user, path, filename, new_name, message):
        """Synchronize, rename a file and propagate.

        :param user: (list)
            VCS user like ``(user_id, password, user_name)``.
        :param path: (string)
            Relative path to files.
        :param filename: (string)
            Name of file to move.
        :param new_name: (string)
            New name.
        :param message: (string)
            Commit message.
        """
        # Check repository
        if not self._check_repository(user, message):
            return

        # Move files
        if self.vcs.rename(path, filename, new_name, self):
            return

        # Propagate
        if not self.vcs.commit_push(message, user[0], user[1], user[2], self):
            self.report('end')

    # -------------------------------------------------------------------------
    def remove(self, user, path, filenames, message):
        """Synchronize, remove files and propagate.

        :param user: (list)
            VCS user like ``(user_id, password, user_name)``.
        :param path: (string)
            Relative path to files.
        :param filenames: (list)
            Names of files to remove.
        :param message: (string)
            Commit message.
        """
        # Check repository
        if not self._check_repository(user, message):
            return

        # Remove files
        for filename in filenames:
            if self.vcs.remove(path, filename, self):
                return

        # Propagate
        if not self.vcs.commit_push(message, user[0], user[1], user[2], self):
            self.report('end')

    # -------------------------------------------------------------------------
    def launch(self, request, action, args=(), kwargs=None):
        """Launch a new action in a thread.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param action: (function)
            Action to launch.
        :param args: (tuple, optional)
            Arguments for the action.
        :param kwargs: (dictionary, optional)
            Keyword arguments for the action.
        :return: (boolean)
            ``True`` if action has been launched.

        Only one action per storage at a time is possible.
        """
        # Is this storage undergoing an action?
        if self._thread is not None and self._thread.is_alive():
            request.session.flash(_('${i}: action already in progress.',
                {'i': self.uid}), 'alert')
            return False
        if self._thread is not None:
            del self._thread

        # Launch action
        kwargs = kwargs or {}
        if action.im_self is not self:
            kwargs['handler'] = self
        self._thread = Thread(
            target=action, name=self.uid, args=args, kwargs=kwargs)
        self._thread.start()
        self.expire = time() + self.cache.expiretime
        return True

    # -------------------------------------------------------------------------
    def _check_repository(self, user, message):
        """Wait for repository availability, synchronize and check push
        capacity.

        :param user: (list)
            VCS user like ``(user_id, password, user_name)``.
        :param message: (string)
            Commit message.
        :return: (boolean)
        """
        k = 10
        while self._report[0] == 'run' and k:
            k -= 1
            sleep(1)
        if self._report[0] == 'run':
            self.report('error', _('storage is busy.'))
            return False

        work = self.synchronize(None, True) and \
            not self.vcs.commit_push(message, user[0], user[1], user[2], self)
        if work:
            return True

        self.vcs.revert_all(self)
        return False



# $Id: site.py 87b7d52fda37 2012/02/23 17:08:44 patrick $
# pylint: disable = I0011, C0322
"""Site view callables."""

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPForbidden

from ..lib.xml import upload_configuration, export_configuration
from ..lib.utils import _, has_permission
from ..lib.views import get_action
from ..lib.form import Form
from ..models import DBSession, close_dbsession
from ..models.users import User
from ..models.groups import Group
from ..models.storages import Storage
from ..models.processors import Processor
from ..models.projects import Project


# =============================================================================
class SiteView(object):
    """Class to manage global website."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='site_admin', renderer='../Templates/sit_admin.pt',
                 permission='sit.adm')
    def admin(self):
        """List users, groups, storages, processors and projects and allow
        import/export operations."""

        action, items = get_action(self._request)
        if action == 'imp!':
            upload_configuration(self._request, 'admin')
            action = ''
        elif action[0:4] == 'exp!':
            return self._export_site(items)

        fbuild = self._request.registry['fbuild']
        fbuild.refresh_agent_list(self._request)

        quantity = {'users': DBSession.query(User).count(),
                    'groups': DBSession.query(Group).count(),
                    'storages': DBSession.query(Storage).count(),
                    'projects': DBSession.query(Project).count()}

        self._request.breadcrumbs.add(_('Site administration'), 2)
        return {
            'form': Form(self._request), 'action': action,
            'db': (DBSession.bind.name, DBSession.bind.url.database),
            'agents': fbuild.agent_list(),
            'processor_labels': sorted(Processor.labels(self._request)),
            'quantity': quantity}

    # -------------------------------------------------------------------------
    def _export_site(self, objects):

        """Export site.
        :param objects: (list)
            List of objects to export.
        :return: (:class:`pyramid.response.Response` instance)
        """
        if not has_permission(self._request, 'admin'):
            raise HTTPForbidden()

        elements = []
        for obj in objects:
            if obj == 'users':
                for user in DBSession.query(User):
                    elements.append(user.xml(True))
            elif obj == 'groups':
                for group in DBSession.query(Group):
                    elements.append(group.xml())
            elif obj == 'storages':
                for storage in DBSession.query(Storage):
                    elements.append(storage.xml())
            elif obj == 'projects':
                for project in DBSession.query(Project):
                    elements.append(project.xml(self._request))

        name = self._request.registry.settings.get('skin.label', 'publiforge')
        if len(objects) == 1:
            name = '%s_%s.pf%s' % (name, objects[0], objects[0][0])
        else:
            name = '%s.pf' % name
        return export_configuration(elements, name)



# $Id: initialize.py ada86c3ea608 2012/03/22 22:52:47 patrick $
"""When you launch the `PubliForge` application, this module determines which
mode you are running: front, agent or both.  According to this mode, it sets up
database, creates a instance of :class:`~.lib.build.front.FrontBuildManager`
class, an instance of :class:`~.lib.handler.HandlerManager`, an instance of
:class:`~.lib.build.agent.AgentBuildManager` class and makes required
directories.
"""

import logging
import sys
import mimetypes
from os import makedirs
from os.path import exists, join, dirname
from shutil import rmtree
from ConfigParser import ConfigParser
from beaker.cache import CacheManager
from beaker.util import parse_cache_config_options
from sqlalchemy import engine_from_config
from sqlalchemy.exc import ProgrammingError

from pyramid.i18n import get_localizer
from pyramid.security import Allow, Authenticated
from pyramid.security import ALL_PERMISSIONS, NO_PERMISSION_REQUIRED
from pyramid.threadlocal import get_current_request
from pyramid.renderers import get_renderer
from pyramid.events import NewRequest, BeforeRender
from pyramid_beaker import session_factory_from_settings

from .lib.utils import _
from .lib.renderer import Breadcrumbs, Menu
from .lib.handler import HandlerManager
from .lib.build.front import FrontBuildManager
from .lib.build.agent import AgentBuildManager
from .models import DBSession, Base
from .models.processors import Processor
from .views.security import connect


LOG = logging.getLogger(__name__)


# =============================================================================
class Initialize(object):
    """Initialization application class."""

    # -------------------------------------------------------------------------
    def __init__(self, global_config, configurator):
        """Constructor method."""
        settings = configurator.registry.settings
        self.configurator = configurator
        self.is_front = bool(settings.get('storage.root'))
        self.is_agent = bool(settings.get('build.root'))
        self._config = ConfigParser({'here': global_config['here']})
        self._config.read(global_config['__file__'])
        self._lang = settings.get('pyramid.default_locale_name')
        mimetypes.init((join(dirname(__file__), 'Static', 'Images',
            'MimeTypes', 'mime.types'),))

    # -------------------------------------------------------------------------
    def all(self):
        """Check settings and create registry objects."""
        # Check general settings
        if not self.is_front and not self.is_agent:
            sys.exit('*** Must be at least a web site ("storage.root" '
                     'needed) or an agent ("build.root" needed).')
        settings = self.configurator.registry.settings
        if not self._lang or not settings.get('available_languages'):
            sys.exit('*** Must define available languages and a default one.')
        if not settings.get('uid'):
            sys.exit('*** Must define an "uid".')

        # Translations
        self.configurator.add_translation_dirs(
            *(settings.get('translation_dirs', '').split()
              + ['publiforge:Locale', 'colander:locale']))

        # Initialize agent
        if self.is_agent:
            self.configurator.registry['abuild'] = AgentBuildManager(settings)
            if len(self.configurator.registry['abuild'].processor_list()) == 0:
                sys.exit('*** Should have defined at least one processor.')
            fronts = self.configurator.registry['abuild'].front_list()
            if len(fronts) == 0:
                sys.exit('*** Must have at least one authorized front.')
            if not settings.get('buildspace.root') \
                   and (len(fronts) > 1 or fronts[0] != settings.get('uid')):
                sys.exit('*** Must define a directory for buildspaces.')
            self.configurator.include(add_routes_agent)

        # Initialize front
        if self.is_front:
            if not settings.get('temporary_dir'):
                sys.exit('*** Must define a temporary directory.')
            if exists(settings['temporary_dir']):
                rmtree(settings['temporary_dir'])
            makedirs(settings['temporary_dir'])
            cache_manager = CacheManager(
                **parse_cache_config_options(settings))
            self.configurator.registry['handler'] = HandlerManager(
                settings, cache_manager)
            if len(self.configurator.registry['handler'].vcs_list()) == 0:
                sys.exit('*** Should have defined at least one VCS.')
            self.configurator.registry['fbuild'] = FrontBuildManager(settings)
            if len(self.configurator.registry['fbuild'].agent_list()) == 0:
                sys.exit('*** Should have at least one available agent.')
            self.configurator.set_session_factory(
                session_factory_from_settings(settings))
            self.configurator.add_subscriber(new_request, NewRequest)
            self.configurator.add_subscriber(before_render, BeforeRender)
            self.configurator.set_default_permission('view')
            self.configurator.include(add_routes_front)
            self.configurator.scan()
            self.skin()
            self.initialize_sql()

        return self.is_front, self.is_agent

    # -------------------------------------------------------------------------
    def skin(self):
        """Set up skin."""
        settings = self.configurator.registry.settings
        if settings.get('skin.static.name'):
            if not settings.get('skin.static.path'):
                sys.exit('*** Must difine a static path.')
            self.configurator.add_static_view(
                settings['skin.static.name'], settings['skin.static.path'],
                cache_max_age=3600)
        if settings.get('skin.static.name') != 'Static':
            self.configurator.add_static_view('Static',
                join(dirname(__file__), 'Static'), cache_max_age=3600)

    # -------------------------------------------------------------------------
    def initialize_sql(self):
        """Database initialization."""
        # Initialize SqlAlchemy session
        settings = self.configurator.registry.settings
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        Base.metadata.bind = engine

        # Check database and empty processor table
        try:
            DBSession.query(Processor).delete()
            DBSession.commit()
        except ProgrammingError:
            sys.exit('*** Run "pfpopulate" script!')
        DBSession.close()


# =============================================================================
def perm_group_finder(user_id, request):
    """Return permission groups of user ``user_id``.

    :param user_id: (string)
        User ID.
    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    """
    # Already authenticated
    if user_id == request.session.get('user_id'):
        return ['group:%s' % perm for perm in request.session['perms']]

    # Auto login
    # pylint: disable = I0011, E1103
    user = connect(request, user_id)
    if not isinstance(user, int):
        user.setup_environment(request)
        return ['group:%s' % perm for perm in request.session['perms']]


# =============================================================================
class RootFactory(object):
    """Return the traversal root of the application.

    Its main role is to define Access Control List (ACL). See
    :ref:`frontreference_permissions` for a complete description of
    permissions.
    """
    # pylint: disable = I0011, R0903

    __acl__ = [
        (Allow, Authenticated, 'view'),
        (Allow, 'group:admin', ALL_PERMISSIONS),
        (Allow, 'group:doc_editor', 'doc.update'),
        (Allow, 'group:doc_manager', ('doc.update', 'doc.create')),
        (Allow, 'group:usr_user', 'usr.read'),
        (Allow, 'group:usr_editor', ('usr.read', 'usr.update')),
        (Allow, 'group:usr_manager', ('usr.read', 'usr.update', 'usr.create')),
        (Allow, 'group:grp_user', 'grp.read'),
        (Allow, 'group:grp_editor', ('grp.read', 'grp.update')),
        (Allow, 'group:grp_manager', ('grp.read', 'grp.update', 'grp.create')),
        (Allow, 'group:stg_user', 'stg.read'),
        (Allow, 'group:stg_editor', ('stg.read', 'stg.update')),
        (Allow, 'group:stg_manager', ('stg.read', 'stg.update', 'stg.create')),
        (Allow, 'group:prj_user', 'prj.read'),
        (Allow, 'group:prj_editor', ('prj.read', 'prj.update')),
        (Allow, 'group:prj_manager', ('prj.read', 'prj.update', 'prj.create')),
        ]

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        pass


# =============================================================================
def new_request(event):
    """A subscriber for :class:`pyramid.events.NewRequest` events."""
    request = event.request
    if not 'lang' in request.session:
        request.session['lang'] = request.accept_language.best_match(
            request.registry.settings.get('available_languages').split(),
            request.registry.settings.get('pyramid.default_locale_name'))
    request._LOCALE_ = request.session['lang']
    request.breadcrumbs = Breadcrumbs(request)


# =============================================================================
def before_render(event):
    """A subscriber for :class:`pyramid.events.BeforeRender` events."""
    # pylint: disable = I0011, W0142
    request = event.get('request') or get_current_request()

    def translate(text, domain='publiforge', **kwargs):
        """Translation from a string."""
        return get_localizer(request).translate(_(text, domain, **kwargs))

    event['base'] = get_renderer(
        request.registry.settings.get('skin.template.base')
        or 'Templates/base.pt').implementation()
    event['title'] = request.registry.settings.get(
        'skin.title', 'PubliForge').decode('utf8')
    event['_'] = translate
    event['route'] = lambda name, *elts, **kwargs: \
        request.route_path(name, *elts, **kwargs).decode('utf8')
    event['Menu'] = Menu


# =============================================================================
def add_routes_front(config):
    """Add Web front routes."""
    # pylint: disable = I0011, R0915
    # Security
    config.add_route('login', '/login')
    config.add_route('login_test', '/login_test')
    config.add_route('logout', '/logout')
    config.add_view('.views.security.login', route_name='login',
        renderer=config.registry.settings.get('skin.template.login')
            or 'Templates/login.pt',
        permission=NO_PERMISSION_REQUIRED)

    # Home
    config.add_route('home', '/')
    config.add_route('favicon', '/favicon.ico')
    config.add_route('robots', '/robots.txt')

    # Site
    config.add_route('site_admin', '/site/admin')

    # User
    config.add_route('user_account', '/user/account')
    config.add_route('user_admin', '/user/admin')
    config.add_route('user_create', '/user/create')
    config.add_route('user_view', '/user/view/{user_id}')
    config.add_route('user_edit', '/user/edit/{user_id}')

    # Group
    config.add_route('group_admin', '/group/admin')
    config.add_route('group_create', '/group/create')
    config.add_route('group_view', '/group/view/{group_id}')
    config.add_route('group_edit', '/group/edit/{group_id}')

    # Storage
    config.add_route('storage_admin', '/storage/admin')
    config.add_route('storage_index', '/storage/index')
    config.add_route('storage_index_ajax', '/storage/index_ajax')
    config.add_route('storage_create', '/storage/create')
    config.add_route('storage_view', '/storage/view/{storage_id}')
    config.add_route('storage_edit', '/storage/edit/{storage_id}')
    config.add_route('storage_root', '/storage/browse/{storage_id}')
    config.add_route('storage_browse',
                     '/storage/browse/{storage_id}/*path')
    config.add_route('storage_root_ajax',
                     '/storage/browse_ajax/{storage_id}')
    config.add_route('storage_browse_ajax',
                     '/storage/browse_ajax/{storage_id}/*path')

    # File
    config.add_route('file_search', '/file/search')
    config.add_route('file_info', '/file/info/{storage_id}/*path')
    config.add_route('file_download', '/file/download/{storage_id}/*path')
    config.add_route('file_revision',
                     '/file/revision/{revision}/{storage_id}/*path')
    config.add_route('file_diff', '/file/diff/{revision}/{storage_id}/*path')

    # Project
    config.add_route('project_admin', '/project/admin')
    config.add_route('project_index', '/project/index')
    config.add_route('project_create', '/project/create')
    config.add_route('project_view', '/project/view/{project_id}')
    config.add_route('project_edit', '/project/edit/{project_id}')
    config.add_route('project_dashboard', '/project/dashboard/{project_id}')
    config.add_route('project_results', '/project/results/{project_id}')
    config.add_route('project_results_ajax',
                     '/project/results_ajax/{project_id}')

    # Processing
    config.add_route('processing_create', '/processing/create/{project_id}')
    config.add_route('processing_view',
                     '/processing/view/{project_id}-{processing_id}')
    config.add_route('processing_edit',
                     '/processing/edit/{project_id}-{processing_id}')

    # Pack
    config.add_route('pack_create', '/pack/create/{project_id}')
    config.add_route('pack_view', '/pack/view/{project_id}-{pack_id}')
    config.add_route('pack_edit', '/pack/edit/{project_id}-{pack_id}')

    # Build
    config.add_route('build_launch',
                     '/build/launch/{project_id}-{processing_id}-{pack_ids}')
    config.add_route('build_view', '/build/view/{build_id}')
    config.add_route('build_progress', '/build/progress/{build_id}')
    config.add_route('build_progress_ajax', '/build/progress_ajax/{build_id}')
    config.add_route('build_complete', '/build/complete/{build_id}/{key}')
    config.add_route('build_log', '/build/log/{build_id}')


# =============================================================================
def add_routes_agent(config):
    """Add agent routes and methods."""
    config.include('pyramid_rpc.xmlrpc')
    config.add_xmlrpc_endpoint('xmlrpc', '/xmlrpc')

    config.add_xmlrpc_method('.views.xmlrpc.processor_list',
        endpoint='xmlrpc', method='processor_list',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.processor_xml',
        endpoint='xmlrpc', method='processor_xml',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method(
        '.views.xmlrpc.activity', endpoint='xmlrpc', method='activity',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method(
        '.views.xmlrpc.start', endpoint='xmlrpc', method='start',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method(
        '.views.xmlrpc.progress', endpoint='xmlrpc', method='progress',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method(
        '.views.xmlrpc.stop', endpoint='xmlrpc', method='stop',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method(
        '.views.xmlrpc.result', endpoint='xmlrpc', method='result',
        permission=NO_PERMISSION_REQUIRED)

    config.add_xmlrpc_method('.views.xmlrpc.buildspace_cleanup',
        endpoint='xmlrpc', method='buildspace_cleanup',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.buildspace_send_signature',
        endpoint='xmlrpc', method='buildspace_send_signature',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.buildspace_receive_file',
        endpoint='xmlrpc', method='buildspace_receive_file',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.buildspace_receive_delta',
        endpoint='xmlrpc', method='buildspace_receive_delta',
        permission=NO_PERMISSION_REQUIRED)

    config.add_xmlrpc_method('.views.xmlrpc.output_list',
        endpoint='xmlrpc', method='output_list',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.output_send_file',
        endpoint='xmlrpc', method='output_send_file',
        permission=NO_PERMISSION_REQUIRED)
    config.add_xmlrpc_method('.views.xmlrpc.output_send_delta',
        endpoint='xmlrpc', method='output_send_delta',
        permission=NO_PERMISSION_REQUIRED)



# $Id: user.py 91a251074d63 2012/03/18 23:07:53 patrick $
# pylint: disable = I0011, C0322
"""User view callables."""

from colander import Mapping, SchemaNode, String, Integer, Date
from colander import All, Length, OneOf, Email, Regex

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPForbidden, HTTPFound, HTTPNotFound

from ..lib.utils import _, has_permission, normalize_name
from ..lib.xml import upload_configuration, export_configuration
from ..lib.views import get_action, paging_users
from ..lib.form import SameAs, Form
from ..lib.renderer import Paging, TabSet
from ..models import ID_LEN, LABEL_LEN, DBSession, close_dbsession
from ..models.users import PERM_SCOPES, USER_PERMS, HOMES, PAGE_SIZE
from ..models.users import USER_STATUS, User, UserPerm, UserIP
from ..models.groups import Group


USER_SETTINGS_TABS = (_('Identity'), _('Permissions'), _('Security'))


# =============================================================================
class UserView(object):
    """Class to manage users."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='user_admin', renderer='../Templates/usr_admin.pt',
                 permission='usr.update')
    def admin(self):
        """List users for administration purpose."""
        action, items = get_action(self._request)
        if action == 'imp!':
            upload_configuration(self._request, 'usr_manager', 'user')
            action = ''
        elif action[0:4] == 'del!':
            self._delete_users(items)
            action = ''
        elif action[0:4] == 'exp!':
            action = self._export_users(items)
            if action:
                return action

        paging, defaults = paging_users(self._request, False)
        form = Form(self._request, defaults=defaults)

        depth = (self._request.breadcrumbs.current_path() ==
                 self._request.route_path('site_admin') and 3) or 2
        self._request.breadcrumbs.add(_('User administration'), depth)
        return {
            'form': form, 'paging': paging, 'action': action,
            'groups': DBSession.query(Group.group_id, Group.label).all(),
            'USER_STATUS': USER_STATUS,
            'i_manager': has_permission(self._request, 'usr_manager')}

    # -------------------------------------------------------------------------
    @view_config(route_name='user_account',
                 renderer='../Templates/usr_view.pt')
    @view_config(route_name='user_view', renderer='../Templates/usr_view.pt')
    def view(self):
        """Show user settings."""
        user_id = self._request.matchdict.get('user_id') \
                  or self._request.session['user_id']
        user_id = int(user_id)
        if user_id != self._request.session['user_id'] \
               and not has_permission(self._request, 'usr_editor'):
            raise HTTPForbidden()

        action = get_action(self._request)[0]
        if action == 'exp!':
            action = self._export_users((user_id,))
            if action:
                return action

        user = DBSession.query(User).filter_by(user_id=user_id).first()
        if user is None:
            raise HTTPNotFound()
        tab_set = TabSet(self._request, USER_SETTINGS_TABS)
        i_editor = has_permission(self._request, 'usr_editor') \
                     or self._request.session['user_id'] == user_id

        self._request.breadcrumbs.add(
            _('User settings'), replace=self._request.route_path('user_edit',
            user_id=user.user_id))
        return {
            'form': Form(self._request), 'tab_set': tab_set,
            'USER_STATUS': USER_STATUS, 'HOMES': HOMES,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
            'user': user, 'i_editor': i_editor}

    # -------------------------------------------------------------------------
    @view_config(route_name='user_create', renderer='../Templates/usr_edit.pt',
                 permission='usr.create')
    def create(self):
        """Create a user."""
        homes = dict([(k, HOMES[k]) for k in HOMES if k != 'site'])
        form, tab_set = self._settings_form()

        if form.validate():
            user = self._create(form.values)
            if user is not None:
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path(
                    'user_edit', user_id=user.user_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        self._request.breadcrumbs.add(_('User creation'))
        return {
            'form': form, 'tab_set': tab_set,
            'USER_STATUS': USER_STATUS, 'homes': homes,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
            'languages': self._request.registry.settings[
                'available_languages'].split(), 'user': None,
            'page_sizes': Paging.page_sizes[1:], 'is_admin': False,
            'is_me': False, 'i_manager': True}

    # -------------------------------------------------------------------------
    @view_config(route_name='user_edit', renderer='../Templates/usr_edit.pt')
    def edit(self):
        """Edit user settings."""
        # Authorization
        user_id = int(self._request.matchdict.get('user_id'))
        if user_id != self._request.session['user_id'] \
               and not has_permission(self._request, 'usr_editor'):
            raise HTTPForbidden()
        user = DBSession.query(User).filter_by(user_id=user_id).first()
        if user is None:
            raise HTTPNotFound()

        # Action
        action = get_action(self._request)[0]
        if action[0:4] == 'rmv!':
            DBSession.query(UserIP).filter_by(
                user_id=user_id, ip=action[4:]).delete()
            DBSession.commit()

        # Environment
        homes = dict([(k, HOMES[k]) for k in HOMES if k != 'site'])
        is_me = user.user_id == self._request.session['user_id']
        i_admin = bool([k for k in user.perms if k.level == 'admin'])
        i_manager = has_permission(self._request, 'usr_manager')
        form, tab_set = self._settings_form(user)

        # Save
        update_perms = i_manager and not i_admin and not is_me
        if action == 'sav!' and form.validate(user) \
               and self._save(user, form.values, update_perms):
            if is_me:
                user.setup_environment(self._request)
            return HTTPFound(self._request.route_path(
                'user_view', user_id=user.user_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs.add(
            _('User settings'), replace=self._request.route_path('user_view',
            user_id=user.user_id))

        return {
            'form': form, 'action': action, 'tab_set': tab_set,
            'USER_STATUS': USER_STATUS, 'homes': homes,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
            'languages': self._request.registry.settings[
                'available_languages'].split(), 'user': user,
            'page_sizes': Paging.page_sizes[1:], 'is_me': is_me,
            'i_admin': i_admin, 'i_manager': i_manager}

    # -------------------------------------------------------------------------
    def _delete_users(self, user_ids):
        """Delete users.

        :param user_ids: (list)
            List of user IDs to delete.
        """
        if not has_permission(self._request, 'usr_manager'):
            raise HTTPForbidden()
        user_ids = set([int(k) for k in user_ids])

        # Do not delete myself
        if self._request.session['user_id'] in user_ids:
            self._request.session.flash(
                _("You can't delete your own user."), 'alert')
            return

        # Do not delete administrators
        if DBSession.query(UserPerm).filter(UserPerm.user_id.in_(user_ids))\
               .filter_by(scope='all', level='admin').first() is not None:
            self._request.session.flash(
                _("You can't delete an administrator."), 'alert')
            return

        # Delete
        DBSession.query(User).filter(
            User.user_id.in_(user_ids)).delete('fetch')
        DBSession.commit()

    # -------------------------------------------------------------------------
    def _export_users(self, user_ids):
        """Export users.

        :param user_ids: (list)
            List of user IDs to export.
        :return: (:class:`pyramid.response.Response` instance or ``''``)
        """
        if not has_permission(self._request, 'usr_user'):
            raise HTTPForbidden()
        i_manager = has_permission(self._request, 'usr_manager')
        elements = []
        for user in DBSession.query(User)\
                .filter(User.user_id.in_(user_ids)).order_by('login'):
            elements.append(user.xml(i_manager))

        elements = [k for k in elements if k is not None]
        if len(elements) == 0:
            self._request.session.flash(
                _('Nothing to do. (You cannot export administrator)'), 'alert')
            return ''

        name = '%s_users.pfu' % self._request.registry.settings.get(
            'skin.label', 'publiforge')
        return export_configuration(elements, name)

    # -------------------------------------------------------------------------
    def _settings_form(self, user=None):
        """Return a user settings form.

        :param user: (:class:`~..models.users.User` instance, optional)
            Current user object.
        :return: (tuple)
            A tuple such as ``(form, tab_set)``.
        """
        # Schema
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='status',
            validator=OneOf(USER_STATUS.keys()), missing='active'))
        schema.add(SchemaNode(String(), name='login',
            validator=Length(min=2, max=ID_LEN)))
        if user is None:
            schema.add(SchemaNode(String(), name='password1', validator=All(
                Length(min=8), SameAs(self._request, 'password2'))))
            schema.add(SchemaNode(String(), name='password2'))
        else:
            schema.add(SchemaNode(String(), name='password1', validator=All(
                Length(min=8), SameAs(self._request, 'password2')),
                missing=None))
            schema.add(SchemaNode(String(), name='password2', missing=None))
        schema.add(SchemaNode(String(), name='name',
            validator=Length(max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='email', validator=All(
            Email(), Length(max=LABEL_LEN))))
        schema.add(SchemaNode(String(), name='lang', validator=OneOf(
            self._request.registry.settings['available_languages'].split())))
        schema.add(SchemaNode(String(), name='home',
            validator=OneOf(HOMES.keys())))
        schema.add(SchemaNode(Integer(), name='page_size',
            validator=OneOf(Paging.page_sizes[1:])))
        schema.add(SchemaNode(Date(), name='expiration', missing=None))
        for scope in PERM_SCOPES:
            schema.add(SchemaNode(String(), name='perm_%s' % scope,
                validator=OneOf(USER_PERMS.keys()), missing='-'))
        schema.add(SchemaNode(String(), name='new_ip',
            validator=Regex(r'(\d{1,3}\.){3}\d{1,3}'), missing=None))

        # Defaults
        if user is not None:
            defaults = {}
            for perm in user.perms:
                defaults['perm_%s' % perm.scope] = perm.level
        else:
            defaults = {'status': 'inactive', 'page_size': PAGE_SIZE,
                        'home': 'projects'}

        return (
            Form(self._request, schema=schema, defaults=defaults, obj=user),
            TabSet(self._request, USER_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _create(self, values):
        """Create a user.

        :param values: (dictionary)
            Form values.
        :return: (:class:`~..models.users.User` instance or ``None``)
        """
        # Check existence
        login = normalize_name(values['login'])[0:ID_LEN]
        user = DBSession.query(User).filter_by(login=login).first()
        if user is not None:
            self._request.session.flash(
                _('This user already exists.'), 'alert')
            return

        # Create user
        # pylint: disable = I0011, W0142
        record = dict([(k, values[k]) for k in values if hasattr(User, k)])
        record['password'] = values['password1']
        if values['new_ip']:
            record['restrict_ip'] = True
        user = User(self._request.registry.settings, **record)
        DBSession.add(user)
        DBSession.commit()

        # Add permissions
        for scope in PERM_SCOPES:
            scope = 'perm_%s' % scope
            if values[scope] != '-':
                user.perms.append(UserPerm(
                    user.user_id, scope[5:], values[scope]))

        # Add IP restriction
        if values['new_ip']:
            user.ips.append(UserIP(user.user_id,  values['new_ip']))

        DBSession.commit()
        return user

    # -------------------------------------------------------------------------
    def _save(self, user, values, update_perms):
        """Save a user settings.

        :param user: (:class:`~..models.users.User` instance)
            User to update.
        :param values: (dictionary)
            Form values.
        :param update_perms: (boolean)
            ``True`` if we can update permissions.
        :return: (boolean)
            ``True`` if succeeds.
        """
        # Update user
        if values['password1']:
            user.set_password(
                self._request.registry.settings, values['password1'])

        # Update permissions
        if update_perms:
            perms = dict([(k.scope, k) for k in user.perms])
            for scope in PERM_SCOPES:
                if values['perm_%s' % scope] != '-':
                    if scope in perms:
                        perms[scope].level = values['perm_%s' % scope]
                    else:
                        user.perms.append(UserPerm(
                            user.user_id, scope, values['perm_%s' % scope]))
                elif scope in perms:
                    DBSession.delete(perms[scope])

        # Add IP restriction
        if values['new_ip'] \
               and values['new_ip'] not in [k.ip for k in user.ips]:
            user.ips.append(UserIP(user.user_id, values['new_ip']))

        DBSession.commit()
        return True



# $Id: users.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
"""SQLAlchemy-powered model definition for users."""

from datetime import datetime
from lxml import etree
from sqlalchemy import Column, ForeignKey, types
from sqlalchemy.orm import relationship

from ..lib.utils import _, hash_sha, normalize_name
from . import ID_LEN, LABEL_LEN, Base, DBSession


USER_STATUS = {'active': _('active'), 'inactive': _('inactive')}
PERM_SCOPES = {'all': _('All'), 'doc': _('Documentation'), 'usr': _('Users'),
    'grp': _('Groups'), 'stg': _('Storages'), 'prj': _('Projects')}
USER_PERMS = {'manager': _('Manager'), 'editor': _('Editor'),
    'user': _('User')}
HOMES = {'site': _('Site administration'), 'projects': _('Project list'),
    'storages': _('Storage list')}
PAGE_SIZE = 20


# =============================================================================
class User(Base):
    """SQLAlchemy-powered user model."""
    # pylint: disable = I0011, R0902, W0142

    __tablename__ = 'users'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    user_id = Column(types.Integer, primary_key=True)
    login = Column(types.String(ID_LEN), unique=True, nullable=False)
    status = Column(types.Enum(*USER_STATUS.keys(), name='usr_status_enum'),
                    nullable=False)
    password = Column(types.String(40))
    name = Column(types.String(LABEL_LEN), nullable=False)
    email = Column(types.String(LABEL_LEN), nullable=False)
    lang = Column(types.String(5), nullable=False)
    expiration = Column(types.Date)
    restrict_ip = Column(types.Boolean, default=False)
    home = Column(types.Enum(*HOMES.keys(), name='usr_home_enum'),
        default='projects', nullable=False)
    page_size = Column(types.Integer, default=PAGE_SIZE)
    last_login = Column(types.DateTime)
    updated = Column(types.DateTime)
    created = Column(types.DateTime)
    perms = relationship('UserPerm')
    ips = relationship('UserIP')

    # -------------------------------------------------------------------------
    def __init__(self, settings, login, status, password, name, email,
                 lang='en', expiration=None, restrict_ip=False, home=None,
                 page_size=None):
        """Constructor method."""
        # pylint: disable = I0011, R0913
        self.login = normalize_name(login)[0:ID_LEN]
        self.status = status
        self.set_password(settings, password)
        self.name = u' '.join(name.split()).strip()[0:LABEL_LEN]
        self.email = email[0:LABEL_LEN]
        self.lang = lang[0:5]
        self.expiration = isinstance(expiration, basestring) \
            and datetime.strptime(expiration, '%Y-%m-%d') or expiration
        self.restrict_ip = restrict_ip
        self.home = home
        self.page_size = page_size
        self.updated = datetime.now()
        self.created = self.updated

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return u"<User({user_id}, '{login}', '{status}', '{name}', '{lang}')>"\
            .format(user_id=self.user_id or -1, login=self.login,
                    status=self.status, name=self.name, lang=self.lang)\
            .encode('utf8')

    # -------------------------------------------------------------------------
    def set_password(self, settings, password):
        """Encrypt and set password.

        :param settings: (dictionary)
            Pyramid deployment settings.
        :param password: (string)
            Clear password.
        """
        if password:
            self.password = hash_sha(password.strip(), settings['auth.key'])

    # -------------------------------------------------------------------------
    def setup_environment(self, request):
        """Set up user environment (session and cookie).

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.

        It saves in session the following values:

        * ``user_id``: user ID
        * ``name``: user name
        * ``page_size``: items per page in listsg
        * ``home``: home page content
        """
        # pylint: disable = I0011, E1101
        request.session['user_id'] = self.user_id
        request.session['name'] = self.name
        request.session['home'] = self.home
        request.session['paging'] = (self.page_size, {})

        # Language
        langs = request.registry.settings.get('available_languages').split()
        request.session['lang'] = (self.lang in langs and self.lang) \
            or (self.lang[0:2] in langs and self.lang[0:2]) \
            or request.registry.settings.get('pyramid.default_locale_name')

        # Permissions
        perms = {}
        levels = {'user': 1, 'editor': 2, 'manager': 3, 'admin': 4}
        for perm in self.perms:
            perms[perm.scope] = perm.level
        for group in self.groups:
            for perm in group.perms:
                if not perm.scope in perms or \
                       levels[perms[perm.scope]] < levels[perm.level]:
                    perms[perm.scope] = perm.level
        if 'all' in perms and perms['all'] != 'admin':
            perm = levels[perms['all']]
            for scope in PERM_SCOPES.keys()[1:]:
                if scope not in perms or levels[perms[scope]] < perm:
                    perms[scope] = perms['all']
            del perms['all']
        request.session['perms'] = perms.get('all') == 'admin' and ('admin',) \
            or tuple(['%s_%s' % (k, perms[k]) for k in perms])

        # Menu
        if 'menu' in request.session:
            del request.session['menu']

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, settings, user_elt, error_if_exists=True):
        """Load a user from a XML file.

        :param settings: (dictionary)
            Application settings.
        :param user_elt: (:class:`lxml.etree.Element` instance)
            User XML element.
        :param error_if_exists: (boolean, default=True)
            It returns an error if user already exists.
        :return: (:class:`pyramid.i18n.TranslationString` or ``None``)
            Error message or ``None``.
        """
        # Check if already exists
        login = normalize_name(user_elt.get('login'))[0:ID_LEN]
        user = DBSession.query(cls).filter_by(login=login).first()
        if user is not None:
            if error_if_exists:
                return _('User "${l}" already exists.', {'l': login})
            return

        # Create user
        # pylint: disable = I0011, W0142
        record = {
            'status': user_elt.get('status', 'active'),
            'password': user_elt.findtext('password') is not None
                 and user_elt.findtext('password').strip() or None,
            'name': u' '.join(user_elt.findtext('name').strip().split()),
            'email': user_elt.findtext('email').strip(),
            'lang': user_elt.findtext('language') is not None
                and user_elt.findtext('language')
                or settings['pyramid.default_locale_name'],
            'expiration': user_elt.findtext('expiration') is not None
                and user_elt.findtext('expiration') or None,
            'restrict_ip': bool(user_elt.find('ips') is not None),
            'home': user_elt.findtext('home') is not None
                and user_elt.findtext('home').strip() or None,
            'page_size': user_elt.findtext('page_size') is not None
                and int(user_elt.findtext('page_size').strip()) or None}
        user = cls(settings, login, **record)
        if record['password'] and user_elt.find('password').get('hash'):
            user.password = record['password']
        DBSession.add(user)
        DBSession.commit()

        # Add permissions
        done = []
        for item in user_elt.findall('permissions/permission'):
            if item.get('scope') not in done:
                user.perms.append(
                    UserPerm(user.user_id, item.get('scope'), item.text))
                done.append(item.get('scope'))

        # Add IP restriction
        done = []
        for item in user_elt.findall('ips/ip'):
            if item.text not in done:
                user.ips.append(UserIP(user.user_id, item.text))
                done.append(item.text)

        DBSession.commit()

    # -------------------------------------------------------------------------
    def xml(self, i_manager=False):
        """Serialize a user to a XML representation.

        :param i_manager: (boolean, default=False)
            Am I a user manager?
        :return: (:class:`lxml.etree.Element`)
        """
        if 'admin' in [k.level for k in self.perms]:
            return

        user_elt = etree.Element('user')
        user_elt.set('login', self.login)
        etree.SubElement(user_elt, 'name').text = self.name
        if i_manager:
            etree.SubElement(
                user_elt, 'password', hash='true').text = self.password
        etree.SubElement(user_elt, 'email').text = self.email
        etree.SubElement(user_elt, 'language').text = self.lang
        etree.SubElement(user_elt, 'home').text = self.home
        if self.page_size != PAGE_SIZE:
            etree.SubElement(user_elt, 'page_size').text = str(self.page_size)
        if self.expiration:
            etree.SubElement(user_elt, 'expiration').text = \
                self.expiration.isoformat()

        # IPs
        if self.ips:
            elt = etree.SubElement(user_elt, 'ips')
            for item in self.ips:
                etree.SubElement(elt, 'ip').text = item.ip

        # Permissions
        if self.perms:
            elt = etree.SubElement(user_elt, 'permissions')
            for perm in self.perms:
                etree.SubElement(
                    elt, 'permission', scope=perm.scope).text = perm.level

        return user_elt


# =============================================================================
class UserPerm(Base):
    """SQLAlchemy-powered user permission class."""
    # pylint: disable = I0011, R0903, W0142

    __tablename__ = 'users_perms'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    user_id = Column(types.Integer,
        ForeignKey('users.user_id', onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True)
    scope = Column(types.Enum(*PERM_SCOPES.keys(), name='perm_scope_enum'),
        primary_key=True)
    level = Column(types.Enum(*(['admin'] + USER_PERMS.keys()),
        name='usr_perms_enum'))

    # -------------------------------------------------------------------------
    def __init__(self, user_id, scope, level):
        """Constructor method."""
        self.user_id = user_id
        self.scope = level == 'admin' and 'all' or scope
        self.level = level

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<UserPerm('{user_id}', '{scope}', '{level}')>".format(
            user_id=self.user_id or -1, scope=self.scope, level=self.level)


# =============================================================================
class UserIP(Base):
    """Class for restriction by IP."""
    # pylint: disable = I0011, C0103, R0903

    __tablename__ = 'users_ips'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    user_id = Column(types.Integer,
        ForeignKey('users.user_id', onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True)
    ip = Column(types.String(40), primary_key=True)

    # -------------------------------------------------------------------------
    def __init__(self, user_id, ip):
        """Constructor method."""
        self.user_id = user_id
        self.ip = ip

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<UserIP({user_id}, '{ip}')>"\
               .format(user_id=self.user_id or -1, ip=self.ip)



# $Id: xmlrpc.py b1ffa5a75b89 2012/02/19 23:23:01 patrick $
"""These XML-RPC functions are called by the front method
:meth:`~.lib.build.front.FrontBuildManager.call`"""

from os import listdir, remove, walk, makedirs, rename
from os.path import exists, join, dirname, relpath, normpath
from shutil import rmtree
from cStringIO import StringIO
from random import randint
import xmlrpclib

from ..lib.utils import _, localizer, camel_case
from ..lib.rsync import get_block_size
from ..lib.rsync import SigFile, PatchedFile, DeltaFile


# =============================================================================
def processor_list(request, context):
    """Return a list of available processors.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :return: (tuple)
        ``(error, processor_list)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)
    return '', request.registry['abuild'].processor_list()


# =============================================================================
def processor_xml(request, context, processor_id):
    """Return XML of processor ``processor_id``.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param processor_id: (string)
        Processor_id
    :return: (tuple)
        ``(<error>, <processor_list>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)
    return '', request.registry['abuild'].processor_xml(processor_id)


# =============================================================================
def activity(request, context):
    """Return the agent activity i.e. the number of active builds.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param password: (string)
        Front password to use agent services.
    :return: (tuple)
        ``(<error>, <activity>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)
    return '', request.registry['abuild'].activity()


# =============================================================================
def start(request, context, build_id, processing, pack, end_url):
    """Start the build.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :param processing: (dictionary)
        A processing dictionary.
    :param pack: (dictionary)
        A pack dictionary.
    :param end_url: (string)
        URL to call to complete the build.
    :return: (tuple)
        ``(<error>, '')``

    Arguments of this RPC function are: ``<front_id>, <password>, <user>,
    <build_id>, <processing>, <pack>, <end_url>``.
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Adjust paths
    if not context['local']:
        processing['resources'] = _agent_paths(
            request, context, *processing['resources'])
        for item in processing['templates']:
            item[0] = _agent_paths(request, context, item[0])[0]
        pack['files'] = _agent_paths(request, context, *pack['files'])
        pack['resources'] = _agent_paths(request, context, *pack['resources'])
        for item in pack['templates']:
            item[0] = _agent_paths(request, context, item[0])[0]

    # Start build
    build_id = _agent_build_id(request, context, build_id)
    request.registry['abuild'].start_build(
        build_id, context, processing, pack, end_url)
    return '', ''


# =============================================================================
def progress(request, context, build_id):
    """Return the progress of a build.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :return: (tuple)
        ``(<error>, <result>)``
    """
    return _agent_method(request, context, build_id, 'progress')


# =============================================================================
def stop(request, context, build_id):
    """Stop a build.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :return: (tuple)
        ``(<error>, <result>)``
    """
    return _agent_method(request, context, build_id, 'stop')


# =============================================================================
def result(request, context, build_id):
    """Return the result of a build.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    """
    return _agent_method(request, context, build_id, 'result')


# =============================================================================
def buildspace_cleanup(request, context, path, filenames):
    """Delete files in path ``path`` of buildspace directory which are not in
    ``files``.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param path: (string)
        Path to clean up.
    :param filenames: (list)
        Names of file to keep.
    :return: (tuple)
        ``(<error>, '')``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Clean directory
    root = join(request.registry.settings['buildspace.root'],
                _agent_paths(request, context, path)[0])
    if not exists(root):
        return '', ''
    for path, dirs, files in walk(root, topdown=False):
        for name in files:
            if not normpath(join(relpath(path, root), name).decode('utf8')) \
                   in filenames and exists(join(path, name)):
                remove(join(path, name))
        for name in dirs:
            if not listdir(join(path, name)):
                rmtree(join(path, name))
    return '', ''


# =============================================================================
def buildspace_send_signature(request, context, filename):
    """Return signature of a buildspace file.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param filename: (string)
        Filename that should send the signature.
    :return: (tuple)
        ``(<error>, <signature>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Empty string or signature
    filename = join(request.registry.settings['buildspace.root'],
                    _agent_paths(request, context, filename)[0])
    if not exists(filename):
        return '', ''
    sig_file = SigFile(open(filename, 'rb'), get_block_size(filename))
    sig_buf = sig_file.read()
    sig_file.close()
    return '', xmlrpclib.Binary(sig_buf)


# =============================================================================
def buildspace_receive_file(request, context, filename, content):
    """Receive a file in buildspace directory.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param filename: (string)
        Filename that should send the signature.
    :param content: (:class:`xmlrpclib.Binary` instance)
        File content.
    :return: (tuple)
        ``(<error>, '')``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Write file
    filename = join(request.registry.settings['buildspace.root'],
                    _agent_paths(request, context, filename)[0])
    if exists(filename):
        return '', ''
    path = dirname(filename)
    if not exists(path):
        makedirs(path)
    with open(filename, 'wb') as hdl:
        hdl.write(content.data)
    return '', ''


# =============================================================================
def buildspace_receive_delta(request, context, filename, delta):
    """Receive a delta (rsync) for a file in buildspace directory.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param filename: (string)
        Filename that should send the signature.
    :param delta: (:class:`xmlrpclib.Binary` instance)
        Patch to apply.
    :return: (tuple)
        ``(<error>, '')``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Patch file
    filename = join(request.registry.settings['buildspace.root'],
                    _agent_paths(request, context, filename)[0])
    if not exists(filename):
        return _error(context, _('Unknown file "${n}"', {'n': filename}))

    patch_file = PatchedFile(open(filename, 'rb'), StringIO(delta.data))
    temp_name = '%s~%d~' % (filename, randint(1, 999999))
    with open(temp_name, 'wb') as hdl:
        hdl.write(patch_file.read())
    patch_file.close()
    if exists(temp_name):
        rename(temp_name, filename)
    return '', ''


# =============================================================================
def output_list(request, context, build_id):
    """List output files of a build.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :return: (tuple)
        ``(<error>, <file_list>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # List files
    build_id = _agent_build_id(request, context, build_id)
    output = join(request.registry.settings['build.root'],
                  camel_case(build_id), 'Output')
    file_list = []
    for path, name, files in walk(output):
        for name in files:
            file_list.append(relpath(join(path, name), output))
    return '', file_list


# =============================================================================
def output_send_file(request, context, build_id, filename):
    """Send a file of the ``Output`` directory.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :param filename: (string)
        Name of the file to transfer.
    :return: (tuple)
        ``(<error>, <file_content>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Send file
    build_id = _agent_build_id(request, context, build_id)
    filename = join(request.registry.settings['build.root'],
                    camel_case(build_id), 'Output', filename)
    with open(filename, 'rb') as hdl:
        return '', xmlrpclib.Binary(hdl.read())


# =============================================================================
def output_send_delta(request, context, build_id, filename, sig):
    """Send a delta list of a file of the ``Output`` directory..

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :param filename: (string)
        Name of the file to transfer.
    :param sig: (:class:`xmlrpclib.Binary` instance)
        File signature.
    :return: (tuple)
        ``(<error>, <delta>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    # Send delta
    build_id = _agent_build_id(request, context, build_id)
    filename = join(request.registry.settings['build.root'],
                    camel_case(build_id), 'Output', filename)
    with open(filename, 'rb') as hdl:
        delta_file = DeltaFile(sig.data, hdl)
        delta_buf = delta_file.read()
        delta_file.close()
    return '', xmlrpclib.Binary(delta_buf)


# =============================================================================
def _authorized(request, context):
    """Check if the request is authorized.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :return: (boolean)
        ``True`` if authorized.
    """
    return request.registry['abuild'].authorized_front(
        context['front_id'], context['password'])


# =============================================================================
def _not_authorized(context):
    """Return ``('Not authozied!', '')`` where error message is translated.

    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :return: (tuple)
    """
    return _error(context, _('Not authorized!'))


# =============================================================================
def _error(context, text):
    """Return ``(text, '')`` where error message ``text`` is translated.

    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param text: (string)
        Error message to trnaslate.
    :return: (tuple)
    """
    return localizer(context['lang']).translate(text), ''


# =============================================================================
def _agent_build_id(request, context, build_id):
    """Prefix ``build_id`` with ``front_id`` if agent is not called locally.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :return: (string)
    """
    if request.registry.settings['uid'] != context['front_id']:
        return '%s-%s' % (context['front_id'].lower(), build_id)
    return build_id


# =============================================================================
def _agent_paths(request, context, *files):
    """Prefix files if agent is not called locally.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param files: (list)
        List of files on front in storage.
    :return: (list)
    """
    if request.registry.settings['uid'] != context['front_id']:
        prefix = camel_case(context['front_id'])
        return ['%s-%s' % (prefix, name) for name in files]
    return files


# =============================================================================
def _agent_method(request, context, build_id, method):
    """Check authentification and return Agent ``method`` result.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param context: (dictionary)
        See :meth:`~.lib.build.front.FrontBuildManager.call`.
    :param build_id: (string)
        Build ID.
    :return: (tuple)
        ``(<error>, <method_result>)``
    """
    if not _authorized(request, context):
        return _not_authorized(context)

    build_id = _agent_build_id(request, context, build_id)
    return '', getattr(request.registry['abuild'], method)(build_id)



# $Id: form.py bc236c725df4 2012/01/24 22:50:00 patrick $
"""Form validation and rendering library."""

import colander
from webhelpers.html import tags, HTML, literal

from pyramid.httpexceptions import HTTPNotAcceptable
from pyramid.i18n import get_localizer

from ..lib.utils import _


# =============================================================================
class SameAs(object):
    # pylint: disable = I0011, R0903
    """This class implements a ``colander`` validator to check if to fields are
    identical."""

    # -------------------------------------------------------------------------
    def __init__(self, request, reference):
        """Constructor method."""
        self.request = request
        self.reference = reference

    # -------------------------------------------------------------------------
    def __call__(self, node, value):
        """This method raises a :class:`colander.Invalid` instance as an
        exception value is not same as ``self.reference``.

        :param node: (:class:`colander.SchemaNode` instance)
        :param value: (cstruct object)
        """
        if self.request.params.get(self.reference) != value:
            raise colander.Invalid(node, _('The two fields are not identical'))


# =============================================================================
def button(url, label='', src=None, title=None, class_='button'):
    """Output a link on a label and an image with a button aspect.

    :param url: (string)
        Target URL.
    :param label: (string, optional)
        Label for roll over and ``alt``.
    :param src: (string, optional)
        Image path.
    :param title: (string, optional)
        Label for roll over.
    :param class_: (string, default='button')
        The class attribute.
    :return: (string)
        HTML tag.
    """
    if class_ == 'button' and not label and src:
        class_ = None
    return literal(u'<a href="{url}"{title}{class_}>{img}{label}</a> '\
        .format(url=url, title=title and ' title="%s"' % title or '',
        class_=class_ and ' class="%s"' % class_ or '',
        img=src and '<img src="%s" alt="%s"/>' % (src, label or title)
        or '', label=label))


# =============================================================================
def grid_item(name, label, content, required=False, hint=None, error=None,
              class_=None, tag='div'):
    """Display an item with label, hint and error message.

    :param name: (string)
        Content ID.
    :param label: (string)
        Label.
    :param content: (string)
        HTML content.
    :param required: (boolean, default=False)
        Indicate if this field is required.
    :param hint: (string, optional)
        Help message.
    :param error: (string, optional)
        Error message.
    :param class_: (string, optional)
        The class attribute.
    :param tag: (string, default='div')
        Tag which contains content, hint and error message.
    :return: (string)

    This ouputs a structure such as:

    .. code-block:: html

        <div class="[class_]">
          <label for="[name]"><em>[label]<strong>*</strong></em></label>
          <tag>
            [content]
            <em>[hint]</em>
            <strong class="error">[form.error(name)]</strong>
          </tag>
          <div class="clear"></div>
        </div>

    If ``class_`` is an empty string, ``'formItem'`` is used.
    """
    # pylint: disable = I0011, R0913
    if not content:
        return ''
    if class_ == '':
        class_ = 'formItem'
    return (literal('<div%s>' % (class_ and ' class="%s"' % class_ or ''))
        + literal('<label%s><em>' % (name and ' for="%s"' % name or ''))
        + label + (HTML.strong('*') if required else '')
        + literal('</em></label><%s>' % tag)
        + (content or literal('&nbsp;'))
        + (HTML.em(' %s' % hint) if hint else '')
        + (HTML.strong(' %s' % error, class_='error') if error else '')
        + literal('</%s><div class="clear"></div></div>' % tag))


# =============================================================================
class Form(object):
    """Form validation class."""
    # pylint: disable = I0011, R0904

    # -------------------------------------------------------------------------
    def __init__(self, request, schema=None, defaults=None, secure=True,
                 obj=None):
        """Constructor method."""
        self.values = not len(request.POST) and defaults or {}
        self._request = request
        self._schema = schema
        self._secure = secure
        self._errors = {}
        self._special = [[], '']
        self._validated = False
        if not len(request.POST) and obj is not None and schema is not None:
            for field in [k.name for k in schema]:
                if hasattr(obj, field.encode('utf8')):
                    self.values[field] = getattr(obj, field)

    # -------------------------------------------------------------------------
    def validate(self, obj=None):
        """Check if the form is validated.

        :param obj: (object, optional)
            Object to fill
        :return: (boolean)
             ``True`` if validated.
        """
        # Something to do?
        if len(self._request.POST) == 0:
            return False
        if self._validated:
            return len(self._errors) == 0

        # Cross-site request forgery protection
        if self._secure and self._request.POST.get('_csrf') \
               != self._request.session.get_csrf_token():
            raise HTTPNotAcceptable()

        # Schema validation
        params = dict(self._request.POST.items())
        if self._schema:
            try:
                self.values = self._schema.deserialize(params)
            except colander.Invalid, err:
                self._errors = {}
                for child in err.children:
                    self._errors[child.node.name] = child.messages()
        else:
            self.values.update(params)

        # Fill object
        if obj is not None and len(self._errors) == 0:
            for field in self.values:
                if hasattr(obj, field.encode('utf8')):
                    setattr(obj, field, self.values[field])

        self._validated = True
        return len(self._errors) == 0

    # -------------------------------------------------------------------------
    def has_error(self, name=None):
        """Return ``True`` if field ``name`` has an error.

        :param name: (string, optional)
            Input ID.
        :return: (boolean)
        """
        return (name is None and bool(self._errors)) or name in self._errors

    # -------------------------------------------------------------------------
    def error(self, name):
        """Return error message for field ``name``.

        :param name: (string)
            Input ID.
        :return: (string)
            Translated error message.
        """
        if not name in self._errors:
            return ''
        translate = get_localizer(self._request).translate
        return ' ; '.join([translate(error) for error in self._errors[name]])

    # -------------------------------------------------------------------------
    def static(self, name):
        """The field ``name`` will not be updated by the form."""
        if name not in self._special[0]:
            self._special[0].append(name)

    # -------------------------------------------------------------------------
    def forget(self, prefix):
        """Fields beginning by ``prefix`` are forgotten when the page is
        refreshed."""
        self._special[1] = prefix[0]

    # -------------------------------------------------------------------------
    def begin(self, url=None, method='post', multipart=False, **attrs):
        """Ouput the ``<form>`` tag.

        :param url: (string, optional)
            URL to submit form, by default, the current URL.
        :param method: (string, default='post')
            The method to use when submitting the form.
        :param multipart: (boolean, default=False)
            If set to ``True``, the enctype is set to ``multipart/form-data``.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        html = tags.form(
            url or self._request.path_qs, method, multipart, **attrs)
        if self._secure:
            token = self._request.session.get_csrf_token() \
                    or self._request.session.new_csrf_token()
            html += HTML.div(self.hidden('_csrf', token))
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def end(cls):
        """Ouput the ``</form>`` tag."""
        return tags.end_form()

    # -------------------------------------------------------------------------
    @classmethod
    def submit(cls, name, label=None, class_='button', **attrs):
        """Output a submit button with the label as the caption.

        :param name: (string)
            Input ID.
        :param label: (string, optional)
            Button caption.
        :param class_: (string, default='button')
            The class attribute.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        return tags.submit(name, label, class_=class_, **attrs)

    # -------------------------------------------------------------------------
    @classmethod
    def submit_image(cls, name, label, src, class_=None):
        """Output an image submit button.

        :param name: (string)
            Input ID.
        :param label: (string)
            Label for roll over and ``alt``.
        :param src: (string)
            Image path.
        :param class_: (string, optional)
            The class attribute.
        :return: (string)
            HTML tag.
        """
        label = label.replace('"', "'")
        return literal(
            u'<input type="image" name="{name}" src="{src}" title="{label}"'
            ' alt="{label}"{class_}/>'.format(name=name, src=src, label=label,
            class_=class_ and ' class="%s"' % class_ or ''))

    # -------------------------------------------------------------------------
    @classmethod
    def submit_cancel(cls, label):
        """Output a cancel submit button.

        :param label: (string)
            Label for roll over and ``alt``.
        :return: (string)
            HTML tag.
        """
        label = label.replace('"', "'")
        return literal(
            u'<input type="image" name="ccl!" '
            'src="/Static/Images/action_cancel.png" title="{label}" '
            'alt="{label}"/>'.format(label=label))

    # -------------------------------------------------------------------------
    @classmethod
    def button(cls, url, label='', src=None, title=None, class_='button'):
        """Output a link on a label and an image with a button aspect.

        See :func:`button`.
        """
        return button(url, label, src, title, class_)

    # -------------------------------------------------------------------------
    @classmethod
    def grid_item(cls, name, label, content, required=False, hint=None,
                  error=None, class_=None, tag='div'):
        """Output an item with label, hint and error messag.

        See :func:`grid_item`.
        """
        # pylint: disable = I0011, R0913
        return grid_item(
            name, label, content, required, hint, error, class_, tag)

    # -------------------------------------------------------------------------
    def hidden(self, name, value=None, **attrs):
        """Output a hidden field.

        :param name: (string)
            Input ID.
        :param value: (string, optional)
            Hidden value.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        return tags.hidden(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def text(self, name, value=None, **attrs):
        """Output a standard text field.

        :param name: (string)
            Input ID.
        :param value: (string, optional)
            Default value.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        return tags.text(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def password(self, name, value=None, **attrs):
        """Output a password field.

        This method takes the same options as text().
        """
        return tags.password(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def checkbox(self, name, value='1', checked=False, label=None, **attrs):
        """Output a check box.

        :param name: (string)
            Input ID.
        :param value: (string, default='1')
            The value to return to the application if the box is checked.
        :param checked: (boolean, default=False)
            ``True`` if the box should be initially checked.
        :param label: (string, optional)
            A text label to display to the right of the box.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        return tags.checkbox(
            name, value, self._value(name, checked), label, **attrs)

    # -------------------------------------------------------------------------
    def select(self, name, selected_values, options, autosubmit=False,
               **attrs):
        """Output a dropdown selection box.

        :param name: (string)
            Input ID.
        :param selected_value: (string or list)
            A string or list of strings or integers giving the value(s) that
            should be preselected.
        :param options: (list of strings, integers or ``(value, label)`` pairs)
            The label will be shown on the form; the option will be returned to
            the application if that option is chosen. If you pass a ``string``
            or ``int`` instead of a ``2-tuple``, it will be used for both the
            value and the label.
        :param autosubmit: (boolean, default=False)
            If ``True``, it adds ``onchange="submit()"`` attribute.
        :param attrs: (dictionary)
            Keyworded arguments for ``webhelpers.html.tags`` object.
        :return: (string)
            HTML tag.
        """
        if isinstance(options[0], tuple):
            translate = get_localizer(self._request).translate
            opts = [(opt[0], translate(opt[1])) for opt in options]
        else:
            opts = options
        return tags.select(name, self._value(name, selected_values), opts,
                    onchange='submit()' if autosubmit else None, **attrs)

    # -------------------------------------------------------------------------
    def upload(self, name, value=None, **attrs):
        """Output a file upload field.

        :param name: (string)
            Input ID.
        :param value: (string, optional)
            Default value.
        :return: (string)
            HTML tag.
        """
        return tags.file(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def grid_text(self, name, label, required=False, hint=None, class_=''):
        """Output a standard text field in a CSS grid layout.

        :param name: (string)
            Input ID.
        :param label: (string)
            Label.
        :param required: (boolean, default=False)
            Indicate if this field is required.
        :param hint: (string, optional)
            Help message.
        :param class_: (string, optional)
            The class attribute.
        :return: (string)
            Output a grid layout.
        """
        return self.grid_item(name, label, self.text(name), required, hint,
                              self.error(name), class_)

    # -------------------------------------------------------------------------
    def grid_password(self, name, label, required=False, hint=None, class_=''):
        """Output a password field in a CSS grid layout.

        This method takes the same options as grid_text().
        """
        return self.grid_item(name, label, self.password(name), required, hint,
                              self.error(name), class_)

    # -------------------------------------------------------------------------
    def grid_checkbox(self, name, label, required=False, hint=None, class_=''):
        """Output a check box in a CSS grid layout.

        :param name: (string)
            Input ID.
        :param label: (string)
            Label.
        :param required: (boolean, default=False)
            Indicate if this field is required.
        :param hint: (string, optional)
            Help message.
        :param class_: (string, optional)
            The class attribute.
        :return: (string)
            Output a grid layout.
        """
        return self.grid_item(name, label, self.checkbox(name), required, hint,
                              self.error(name), class_, tag='span')

    # -------------------------------------------------------------------------
    def grid_select(self, name, label, options, autosubmit=False,
                    required=False, hint=None, class_=''):
        """Output a dropdown selection box in a CSS grid layout.

        :param name: (string)
            Input ID.
        :param label: (string)
            Label.
        :param options: (list of strings, integers or ``(value, label)`` pairs)
            Values in the dropdown list.
        :param autosubmit: (boolean, default=False)
            If ``True``, it adds ``onchange="submit()"`` attribute.
        :param required: (boolean, default=False)
            Indicate if this field is required.
        :param hint: (string, optional)
            Help message.
        :param class_: (string, optional)
            The class attribute.
        :return: (string)
            Output a grid layout.
        """
        # pylint: disable = I0011,R0913
        return self.grid_item(
            name, label, self.select(name, None, options, autosubmit),
            required, hint, self.error(name), class_)

    # -------------------------------------------------------------------------
    def grid_upload(self, name, label, required=False, hint=None, class_=''):
        """Output a file upload field in a CSS grid layout.

        :param name: (string)
            Input ID.
        :param label: (string)
            Label.
        :param required: (boolean, default=False)
            Indicate if this field is required.
        :param hint: (string, optional)
            Help message.
        :param class_: (string, optional)
            The class attribute.
        :return: (string)
            Output a grid layout.
        """
        return self.grid_item(name, label, self.upload(name), required, hint,
                              self.error(name), class_)

    # -------------------------------------------------------------------------
    def _value(self, name, default=None):
        """Return the best value for the field ``name``.

        :param name: (string)
            Input ID.
        :param default: (string, optional)
            Default value.
        """
        return (name not in self._special[0] and name[0] != self._special[1]
            and name in self._request.params and self._request.params[name]) \
            or (name in self.values and self.values[name]) or default



# $Id: processors.py 686a5b4faf1e 2012/01/05 22:44:46 patrick $
"""SQLAlchemy-powered model definition for processors."""

from time import time
from os.path import join, dirname
from cStringIO import StringIO
from lxml import etree

from sqlalchemy import Column, types

from ..lib.xml import local_text, load
from . import Base, DBSession, ID_LEN


# =============================================================================
class Processor(Base):
    """SQLAlchemy-powered processor model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'processors'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    processor_id = Column(types.String(ID_LEN), primary_key=True)
    xml = Column(types.Text())
    updated = Column(types.Float())

    # -------------------------------------------------------------------------
    def __init__(self, processor_id):
        """Constructor method."""
        self.processor_id = processor_id.strip()[0:ID_LEN]
        self.updated = time()

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<Processor('{processor_id}')>".format(
            processor_id=self.processor_id)

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, processor_id, xml):
        """Load a processor from a XML string.

        :param processor_id: (string)
            Processor ID.
        :param xml: (string)
            XML string.
        :return: (string, :class:`pyramid.i18n.TranslationString` or
            :class:`Processor`)
            Error message or processor object.
        """
        if isinstance(xml, unicode):
            xml = xml.encode('utf8')
        tree = load('processor.xml', {'publiforge':
            join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')}, xml)
        if isinstance(tree, basestring):
            return tree

        # Check if already exists
        processor = DBSession.query(cls).filter_by(
            processor_id=processor_id).first()
        if processor is None:
            processor = cls(processor_id)

        processor.xml = xml.decode('utf8')
        processor.updated = time()

        return processor

    # -------------------------------------------------------------------------
    @classmethod
    def labels(cls, request):
        """Labels of all processors in local language or default language.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :return: (dictionary)
            A dictionary such as ``{<processor_id>:
            <processor_label>,...}``.

        If some labels do not exist in the asked language, this method returns
        the label in default language or ``processor_id``.
        """
        request.registry['fbuild'].refresh_agent_list(request)
        labels = {}
        for processor in DBSession.query(cls.processor_id, cls.xml):
            tree = etree.parse(StringIO(processor[1].encode('utf8')))
            labels[processor[0]] = local_text(
                tree.find('processor'), 'label', request, default=processor[0])

        return labels

    # -------------------------------------------------------------------------
    @classmethod
    def description(cls, request, processor_id=None, xml=None):
        """Description of processor ``processor_id`` in local language or
        default language.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param processor_id: (string, optional)
            Processor ID.
        :param xml: (:class:`lxml.etree.ElementTree` instance, optional)
            Processor XML.
        :return: (string)
            Description.

        If some labels do not exist in the asked language, this method returns
        the label in default language or ``processor_id``.
        """
        if processor_id is not None:
            request.registry['fbuild'].refresh_agent_list(request)
            xml = DBSession.query(cls.xml).filter_by(
                processor_id=processor_id).first()
            if xml is not None:
                xml = etree.parse(StringIO(xml[0].encode('utf8')))
        if xml:
            return local_text(xml.find('processor'), 'description', request)



# $Id: __init__.py a8f1b45b2ed8 2011/12/10 10:19:48 patrick $
"""Provides a high-level interface to some librsync functions.

This is a python wrapper around the lower-level _librsync module,
which is written in C.
"""

import types
import array
from os.path import getsize

from ...lib.rsync import _librsync

BLOCKSIZE = _librsync.RS_JOB_BLOCKSIZE


# =============================================================================
class LikeFile(object):
    """File-like object used by SigFile, DeltaFile, and PatchFile."""

    # This will be replaced in subclasses by an object with
    # appropriate cycle() method
    maker = None
    mode = 'rb'

    # -------------------------------------------------------------------------
    def __init__(self, infile, need_seek=None):
        """LikeFile initializer - zero buffers, set eofs off."""
        self.check_file(infile, need_seek)
        self.infile = infile
        self.closed = self.infile_closed = None
        self.inbuf = ""
        self.outbuf = array.array('c')
        self.eof = self.infile_eof = None

    # -------------------------------------------------------------------------
    @classmethod
    def check_file(cls, likefile, need_seek=None):
        """Raise type error if ``likefile`` doesn't have necessary attributes.
        """
        if not hasattr(likefile, 'read'):
            raise TypeError('Basis file must have a read() method')
        if not hasattr(likefile, 'close'):
            raise TypeError('Basis file must have a close() method')
        if need_seek and not hasattr(likefile, 'seek'):
            raise TypeError('Basis file must have a seek() method')

    # -------------------------------------------------------------------------
    def read(self, length=-1):
        """Build up self.outbuf, return first length bytes."""
        if length == -1:
            while not self.eof:
                self._add_to_outbuf_once()
            real_len = len(self.outbuf)
        else:
            while not self.eof and len(self.outbuf) < length:
                self._add_to_outbuf_once()
            real_len = min(length, len(self.outbuf))

        return_val = self.outbuf[:real_len].tostring()
        del self.outbuf[:real_len]
        return return_val

    # -------------------------------------------------------------------------
    def _add_to_outbuf_once(self):
        """Add one cycle's worth of output to self.outbuf."""
        if not self.infile_eof:
            self._add_to_inbuf()
        try:
            self.eof, len_inbuf_read, cycle_out = self.maker.cycle(self.inbuf)
        except _librsync.librsyncError, err:
            raise LibRSyncError(str(err))
        self.inbuf = self.inbuf[len_inbuf_read:]
        self.outbuf.fromstring(cycle_out)

    # -------------------------------------------------------------------------
    def _add_to_inbuf(self):
        """Make sure len(self.inbuf) >= BLOCKSIZE."""
        assert not self.infile_eof
        while len(self.inbuf) < BLOCKSIZE:
            new_in = self.infile.read(BLOCKSIZE)
            if not new_in:
                self.infile_eof = 1
                assert not self.infile.close()
                self.infile_closed = 1
                break
            self.inbuf += new_in

    # -------------------------------------------------------------------------
    def close(self):
        """Close infile."""
        if not self.infile_closed:
            assert not self.infile.close()
        self.closed = 1


# =============================================================================
class SigFile(LikeFile):
    """File-like object which incrementally generates a librsync signature"""

    # -------------------------------------------------------------------------
    def __init__(self, infile, blocksize=BLOCKSIZE):
        """SigFile initializer - takes basis file.

        basis file only needs to have read() and close() methods.  It
        will be closed when we come to the end of the signature.

        """
        LikeFile.__init__(self, infile)
        try:
            self.maker = _librsync.new_sigmaker(blocksize)
        except _librsync.librsyncError, err:
            raise LibRSyncError(str(err))


# =============================================================================
class DeltaFile(LikeFile):
    """File-like object which incrementally generates a librsync delta."""

    # -------------------------------------------------------------------------
    def __init__(self, signature, new_file):
        """DeltaFile initializer - call with signature and new file

        Signature can either be a string or a file with read() and
        close() methods.  New_file also only needs to have read() and
        close() methods.  It will be closed when self is closed.

        """
        LikeFile.__init__(self, new_file)
        if type(signature) is types.StringType:
            sig_string = signature
        else:
            self.check_file(signature)
            sig_string = signature.read()
            assert not signature.close()
        try:
            self.maker = _librsync.new_deltamaker(sig_string)
        except _librsync.librsyncError, err:
            raise LibRSyncError(str(err))


# =============================================================================
class PatchedFile(LikeFile):
    """File-like object which applies a librsync delta incrementally."""

    # -------------------------------------------------------------------------
    def __init__(self, basis_file, delta_file):
        """PatchedFile initializer - call with basis delta

        Here basis_file must be a true Python file, because we may
        need to seek() around in it a lot, and this is done in C.
        delta_file only needs read() and close() methods.

        """
        LikeFile.__init__(self, delta_file)
        if type(basis_file) is not types.FileType:
            raise TypeError("basis_file must be a (true) file")
        try:
            self.maker = _librsync.new_patchmaker(basis_file)
        except _librsync.librsyncError, err:
            raise LibRSyncError(str(err))


# =============================================================================
class LibRSyncError(Exception):
    """Signifies error in internal librsync processing (bad signature, etc.)

    underlying _librsync.librsyncError's are regenerated using this
    class because the C-created exceptions are by default
    unPickleable.  There is probably a way to fix this in _librsync,
    but this scheme was easier.
    """
    pass


# =============================================================================
def get_block_size(filename):
    """
    Return a reasonable block size to use on files of length file_len

    If the block size is too big, deltas will be bigger than is
    necessary. If the block size is too small, making deltas and
    patching can take a really long time.
    """
    file_len = getsize(filename)
    if file_len < 1024000:
        return 512  # set minimum of 512 bytes
    else:
        # Split file into about 2000 pieces, rounding to 512
        file_blocksize = long((file_len / (2000 * 512)) * 512)
        return min(file_blocksize, 2048L)



# $Id: test_views_security.py ffe4e2582c93 2011/12/18 21:51:32 patrick $
# pylint: disable = I0011, R0904
"""Tests of ``views.security`` functions."""

from ..tests import FunctionalTestCase


# =============================================================================
class FunctionalTestViewsSecurityLogin(FunctionalTestCase):
    """Functional test class for ``views.security.login``."""
    # pylint: disable = I0011, C0103, R0904

    # -------------------------------------------------------------------------
    def __init__(self, method_name='runTest'):
        """Constructor method."""
        super(FunctionalTestViewsSecurityLogin, self).__init__(method_name)
        self._csrf = None

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        FunctionalTestCase.setUp(self)
        response = self.testapp.get('/login')
        self._csrf = response.form.get('_csrf').value

    # -------------------------------------------------------------------------
    def test_empty_password(self):
        """[f:views.security.login] empty password."""
        params = {'_csrf': self._csrf, 'login': 'user1', 'password': None}
        response = self.testapp.post('/login', params, status=200)
        self.assertTrue('class="error"' in response)

    # -------------------------------------------------------------------------
    def test_wrong_password(self):
        """[f:views.security.login] wrong password."""
        params = {'_csrf': self._csrf, 'login': 'user1',
                  'password': '12345678'}
        response = self.testapp.post('/login', params, status=200)
        self.assertTrue('"flashAlert"' in response)
        self.assertTrue('incorrect' in response)

    # -------------------------------------------------------------------------
    def test_inactive(self):
        """[f:views.security.login] inactive account."""
        params = {'_csrf': self._csrf, 'login': 'test1',
                  'password': 'test1pwd'}
        response = self.testapp.post('/login', params, status=200)
        self.assertTrue('"flashAlert"' in response)
        self.assertTrue('activ' in response)

    # -------------------------------------------------------------------------
    def test_expired(self):
        """[f:views.security.login] inactive expired."""
        params = {'_csrf': self._csrf, 'login': 'test3',
                  'password': 'test3pwd'}
        response = self.testapp.post('/login', params, status=200)
        self.assertTrue('"flashAlert"' in response)
        self.assertTrue('expir' in response)

    # -------------------------------------------------------------------------
    def test_authorized(self):
        """[f:views.security.login] authorized user."""
        from datetime import datetime, timedelta
        from ..models import DBSession
        from ..models.users import User
        params = {'_csrf': self._csrf, 'login': 'user1',
                  'password': 'user1pwd'}
        self.testapp.post('/login', params, status=302)
        user = DBSession.query(User).filter_by(login=params['login']).first()
        self.assertNotEqual(user, None)
        self.assertTrue(
            user.last_login > datetime.now() - timedelta(minutes=1))



# $Id: home.py bc236c725df4 2012/01/24 22:50:00 patrick $
# pylint: disable = I0011, C0322
"""Home view callables."""

from os.path import join, dirname

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound
from pyramid.response import Response
from pyramid.asset import abspath_from_asset_spec
from pyramid.security import NO_PERMISSION_REQUIRED

from ..lib.utils import _
from ..models import close_dbsession


# =============================================================================
class HomeView(object):
    """Class to manage *Home* pages."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='home', renderer='../Templates/index.pt')
    def index(self):
        """Index page."""
        if 'breadcrumbs' in self._request.session:
            del self._request.session['breadcrumbs']

        home = self._request.session.get('home')
        if home == 'projects':
            return HTTPFound(self._request.route_path('project_index'))
        elif  home == 'storages':
            return HTTPFound(self._request.route_path('storage_index'))
        elif  home == 'site':
            return HTTPFound(self._request.route_path('site_admin'))

        self._request.breadcrumbs.add(_('Home'), 1, 1)
        return {}


# =============================================================================
@view_config(route_name='favicon', permission=NO_PERMISSION_REQUIRED)
def favicon(request):
    """Output the `favicon.ico` file."""
    settings = request.registry.settings
    if settings.get('skin.favicon') and settings.get('skin.static.name'):
        icon = join(abspath_from_asset_spec(settings['skin.static.path']),
                    settings.get('skin.favicon')[1:].partition('/')[2])
    else:
        icon = join(dirname(__file__), '..', 'Static', 'favicon.ico')
    with open(icon, 'rb') as hdl:
        body = hdl.read()
    return Response(content_type='image/x-icon', body=body)


# =============================================================================
@view_config(route_name='robots', permission=NO_PERMISSION_REQUIRED)
def robots(request):
    """Output the `robots.txt` file."""
    # pylint: disable = I0011, W0613
    with open(join(dirname(__file__), '..', 'Static', 'robots.txt')) as hdl:
        body = hdl.read()
    return Response(content_type='text/plain', body=body)



# $Id: hg.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
"""Storage with Mercurial Version Control management."""

import logging
from os import remove, renames, makedirs
from os.path import exists, join, isdir, getmtime
from shutil import rmtree
from datetime import datetime
from urllib2 import HTTPError
from mercurial import ui, hg, commands
from mercurial.error import Abort
from urllib2 import URLError
from tempfile import mkdtemp

from pyramid.i18n import TranslationString

from ...lib.utils import _
from ...lib.vcs import Vcs


LOG = logging.getLogger(__name__)


# =============================================================================
class VcsMercurialUi(ui.ui):
    """Override Mercurial ui class for Pylons use."""
    # pylint: disable = I0011, R0904

    # -------------------------------------------------------------------------
    def __init__(self, src=None):
        """Contructor method."""
        super(VcsMercurialUi, self).__init__(src)
        self.handler = None

    # -------------------------------------------------------------------------
    def write(self, *args, **opts):
        """Write args in log as informations."""
        for arg in args:
            LOG.info(arg.strip())

    # -------------------------------------------------------------------------
    def write_err(self, *args, **opts):
        """Write args in log as errors."""
        for arg in args:
            LOG.error(arg.strip())
            if self.handler is not None:
                self.handler.report('error', arg.strip())


# =============================================================================
class VcsMercurial(Vcs):
    """Version control system with Mercurial."""

    engine = 'hg'

    # -------------------------------------------------------------------------
    def __init__(self, path, url, user_id=None, password=None):
        """Constructor method."""
        super(VcsMercurial, self).__init__(path, url, user_id, password)
        self._ui = VcsMercurialUi()
        self._ui.setconfig('ui', 'interactive', 'no')
        self._ui.setconfig('ui', 'username', '-')
        self._ui.setconfig('web', 'cacerts', '')

    # -------------------------------------------------------------------------
    def clone(self, handler=None):
        """Create a copy of an existing repository in a directory.

        See abstract function :meth:`~.lib.vcs.Vcs.clone`.
        """
        self._ui.handler = handler
        try:
            commands.clone(self._ui, self._full_url(), self.path)
        except (Abort, HTTPError, URLError, OSError), error:
            if handler is not None:
                handler.report('error', error)
                if not exists(self.path):
                    makedirs(self.path)
            return error
        remove(join(self.path, '.hg', 'hgrc'))

    # -------------------------------------------------------------------------
    def pull_update(self, handler):
        """Pull changes and update.

        See abstract function :meth:`~.lib.vcs.Vcs.pull_update`.
        """
        if not exists(self.path):
            return
        try:
            if self.engine == 'local':
                commands.update(self._ui, self._repo(handler))
            else:
                commands.pull(self._ui, self._repo(handler),
                                     self._full_url(), update=True)
        except (Abort, HTTPError, URLError, OSError), error:
            return handler.report('error', error)

    # -------------------------------------------------------------------------
    def commit_push(self, message, user_id, password, name, handler):
        """Commit and push changes.

        See abstract function :meth:`~.lib.vcs.Vcs.commit_push`.
        """
        if not exists(self.path):
            return
        my_ui = VcsMercurialUi()
        my_ui.handler = handler
        my_ui.setconfig('ui', 'interactive', 'no')
        my_ui.setconfig('ui', 'username', name.encode('utf8') or user_id)
        my_ui.setconfig('web', 'cacerts', '')
        repo = hg.repository(my_ui, self.path)
        try:
            commands.commit(my_ui, repo, message=message.encode('utf8'))
            if self.engine != 'local':
                commands.push(my_ui, repo, self._full_url(user_id, password))
        except (Abort, HTTPError, URLError, OSError), error:
            return handler.report('error', error)

    # -------------------------------------------------------------------------
    def revert_all(self, handler):
        """Revert all files of the repository."""
        if not exists(self.path):
            return
        commands.revert(
            self._ui, self._repo(handler), all=True, no_backup=True)
        commands.update(self._ui, self._repo(handler), clean=True)

    # -------------------------------------------------------------------------
    def last_change(self):
        """Return the last change on the repository.

        See :meth:`~.lib.vcs.Vcs.last_change`.
        """
        ctx = self._repo()['tip']
        if ctx.rev() == -1:
            return (datetime.fromtimestamp(getmtime(self.path)), -1, '-')
        return (datetime.fromtimestamp(ctx.date()[0]), ctx.rev(),
                ctx.user().partition('<')[0].decode('utf8'))

    # -------------------------------------------------------------------------
    def file_infos(self, path, filename):
        """Return information on a file.

        See :meth:`~.lib.vcs.Vcs.last_change`.
        """
        log = self.log(path, filename)
        if log is None:
            return
        return len(log) and log[0] \
            or (datetime.fromtimestamp(
                getmtime(self.full_path(path, filename))), '-', '-', '-')

    # -------------------------------------------------------------------------
    def log(self, path, filename, limit=1):
        """show revision history of file ``filename``.

        See :meth:`~.lib.vcs.Vcs.log`.
        """
        filename = self._log_path(path, filename)
        if filename is None:
            return

        class Ui(VcsMercurialUi):
            """Ui class to retrieve file information."""
            # pylint: disable = I0011, R0904
            def __init__(self, src=None):
                super(Ui, self).__init__(src)
                self.log = []
                self._entry = None

            def write(self, *args, **opts):
                for arg in args:
                    if arg.startswith('changeset:'):
                        self._entry = [
                            '', int(arg[10:].partition(':')[0]), '', '']
                    elif arg.startswith('user:'):
                        self._entry[2] = arg[5:].strip()
                    elif arg.startswith('date:'):
                        self._entry[0] = arg[5:-6].strip()
                    elif arg.startswith('summary:'):
                        self._entry[3] = arg[8:].strip()
                        self.log.insert(0, self._entry)

        my_ui = Ui()
        my_ui.setconfig('ui', 'verbose', False)
        repo = hg.repository(my_ui, self.path)
        try:
            commands.log(my_ui, repo, filename, limit=str(limit),
                follow=limit > 1 and not isdir(filename),
                date=None, rev=None, user=None)
        except (Abort, HTTPError, URLError, OSError), error:
            return [(datetime.now(), '', '', error)]
        if not len(my_ui.log):
            return my_ui.log

        my_ui.log = sorted(my_ui.log, key=lambda k: k[1], reverse=True)[:limit]
        for k, entry in enumerate(my_ui.log):
            my_ui.log[k] = (
                datetime.strptime(entry[0], '%a %b %d %H:%M:%S %Y'),
                str(entry[1]), entry[2].partition('<')[0].decode('utf8'),
                entry[3].decode('utf8'))
        return my_ui.log

   # -------------------------------------------------------------------------
    def add(self, path, handler):
        """Add all new files in ``path``.

        See abstract function :meth:`~.lib.vcs.Vcs.add`.
        """
        path = self.full_path(path)
        if isinstance(path, TranslationString):
            return handler.report('error', path)
        if not exists(path):
            return
        warn = commands.add(self._ui, self._repo(handler), path)
        if warn:
            return handler.report('error', 'Rejected')

    # -------------------------------------------------------------------------
    def rename(self, path, filename, new_name, handler):
        """Rename a file."""
        filename = self.full_path(path, filename)
        if isinstance(filename, TranslationString):
            return handler.report('error', filename)
        new_name = self.full_path(path, new_name)
        if isinstance(new_name, TranslationString):
            return handler.report('error', new_name)
        if exists(new_name):
            return handler.report('error', _('File already exists!'))
        try:
            commands.rename(self._ui, self._repo(handler), filename, new_name)
        except (HTTPError, URLError, OSError), error:
            return handler.report('error', error)
        except Abort, error:
            if exists(filename) and not exists(new_name):
                renames(filename, new_name)

    # -------------------------------------------------------------------------
    def remove(self, path, filename, handler):
        """Remove a file.

        See abstract function :meth:`~.lib.vcs.Vcs.remove`.
        """
        filename = self.full_path(path, filename)
        if isinstance(filename, TranslationString):
            return handler.report('error', filename)
        if not exists(filename):
            return
        try:
            commands.remove(
                self._ui, self._repo(handler), filename)
        except (Abort, HTTPError, URLError, OSError), error:
            return handler.report('error', error)
        if exists(filename):
            if isdir(filename):
                rmtree(filename)
            else:
                remove(filename)

    # -------------------------------------------------------------------------
    def revision(self, fullname, revision):
        """Retrieve a revision.

        See :meth:`~.lib.vcs.Vcs.revision`.
        """
        tmp_dir = mkdtemp(prefix='publiforge')
        tmp_fil = join(tmp_dir, 'output')
        err = commands.cat(
            self._ui, self._repo(), fullname, rev=revision, output=tmp_fil)
        if err:
            rmtree(tmp_dir)
            return
        with open(tmp_fil, 'rb') as hdl:
            content = hdl.read()
        rmtree(tmp_dir)
        return content

    # -------------------------------------------------------------------------
    def diff(self, fullname, revision):
        """Return differences between revision ``revision`` and current
        revision.

        See :meth:`~.lib.vcs.Vcs.diff`.
        """
        class Ui(VcsMercurialUi):
            """Ui class to retrieve file information."""
            # pylint: disable = I0011, R0904
            def __init__(self, src=None):
                super(Ui, self).__init__(src)
                self.diff = ''

            def write(self, *args, **opts):
                self.diff += ''.join(args)

        my_ui = Ui()
        repo = hg.repository(my_ui, self.path)
        try:
            commands.diff(my_ui, repo, fullname, rev=[revision])
        except RuntimeError, err:
            return err

        return my_ui.diff.decode('utf8')

    # -------------------------------------------------------------------------
    def _repo(self, handler=None):
        """Get a repository object.

        :param handler: (:class:`~.lib.handler.Handler` instance, optional)
            Owner of this action.
        :return: (:class:`mercurial.repo`)
        """
        if handler is not None:
            self._ui.handler = handler
        return hg.repository(self._ui, self.path)


# =============================================================================
class VcsLocal(VcsMercurial):
    """Version control system for local files."""

    engine = 'local'

    # -------------------------------------------------------------------------
    def clone(self, handler=None):
        """Initialize a Mercurial repository and copy source.

        See abstract function :meth:`~.lib.vcs.Vcs.clone`.
        """
        self._ui.handler = handler
        commands.init(self._ui, self.path)
        repo = self._repo()
        commands.add(self._ui, repo)
        commands.commit(self._ui, repo, message='Initial commit')



# $Id: transform.py 816eec7f0a0c 2012/03/22 10:22:19 patrick $
"""XML Transformation via XSL stylesheet."""

from os import walk, remove
from os.path import exists, join, dirname, relpath, getmtime, commonprefix
from lxml import etree
import re
import fnmatch

from ...utils import _, camel_case
from ...xml import load
from . import containers
from .iniscript import IniScript


# =============================================================================
class Transform(object):
    """Class for XML transformation."""
    # pylint: disable = I0011, E1103, R0902

    # -------------------------------------------------------------------------
    def __init__(self, engine):
        """Constructor method.

        :param engine: (:class:`~.lib.engine.leprisme.Engine` instance)
            Engine object on which it depends.
        """
        self._engine = engine
        self.fid = None
        self.data = None
        self.output = None

        # Load regex and XSL
        self._regex = self._load_regex()
        self._xslt = self._load_xslt()

        # List container factories
        self._factories = {
            'Zip': containers.ZipFactory(engine),
            'OCF': containers.OcfFactory(engine)}

        # INI script
        self._iniscript = IniScript(engine)

    # -------------------------------------------------------------------------
    def start(self, filename, fid, data, mtime=None):
        """Start the transformation.

        :param filename: (string)
            Relative path to the original file to transform.
        :param fid: (string)
            File ID.
        :param data: (string or :class:`lxml.etree.ElementTree` instance)
            Name of file or tree to transform.
        :param mtime: (timestamp, optional)
            Time of last modification of the source file.
        """
        # pylint: disable = I0011, R0911
        # Initialization
        if data is None or self._engine.build.stopped():
            return
        self.fid = fid
        self.data = data
        self.output = self._initialize(filename, mtime)
        if self.output is None or self._engine.build.stopped():
            return
        self._save_tmp(0)

        # Preprocess
        self._script_transform('preprocess', filename)
        if self.data is None or self._engine.build.stopped():
            return

        # Pre regex transformation
        self._regex_transform('preregex')
        if self.data is None or self._engine.build.stopped():
            return

        # XSL transformation
        self._xsl_transform(self._engine.build.processing['variables'])
        if self.data is None or self._engine.build.stopped():
            return

        # Generated INI files
        self._ini_scripts(filename)
        if self._engine.build.stopped():
            return

        # Post regex transformation
        self._regex_transform('postregex')
        if self.data is None or self._engine.build.stopped():
            return

        # Postprocess
        self._script_transform('postprocess', filename)

        # Finalize
        self._finalize()

    # -------------------------------------------------------------------------
    def _load_regex(self):
        """Load regular expressions from file.

        :return: (dictionary)
            A dictionary of 2 list of regular expressions.
        """
        regex = {}
        for kind in ('preregex', 'postregex'):
            # Load file
            name = self._engine.config('Transformation', kind)
            if not name:
                continue
            if not exists(name):
                self._engine.build.stopped(
                    'Unknown file "%s".' % name)
                return regex

            # Browse file
            regex[kind] = []
            for line in open(name, 'r'):
                line = line.strip()
                if line and not line[0] == '#' and not line[0:7] == '[Regex]':
                    line = line.partition(' =')
                    repl = line[2].strip()
                    if repl and repl[0] in '\'"' and repl[-1] in '\'"':
                        repl = repl[1:-1]
                    if repl.startswith('lambda'):
                        repl = eval(repl)
                    regex[kind].append(
                        (re.compile(line[0].strip(), re.MULTILINE), repl))
        return regex

    # -------------------------------------------------------------------------
    def _load_xslt(self):
        """Load XSL file and create a etree.XSLT object.

        :return: (:class:`lxml.etree.XSLT` instance)
        """
        filename = self._engine.config('Transformation', 'xsl')
        if not filename:
            return

        try:
            _xslt = etree.XSLT(etree.parse(filename))
        except (IOError, etree.XSLTParseError, etree.XMLSyntaxError), err:
            self._engine.build.stopped(
                str(err).replace(self._engine.build.path, '..'))
            return
        return _xslt

    # -------------------------------------------------------------------------
    def _initialize(self, name, mtime=None):
        """Prepare output directory and compute target name.

        :param name: (string)
            Relative path to file to transform.
        :param mtime: (timestamp, optinal)
            Time of last modification of the source file.
        :return: (string)
            Full path to output directory.
        """
        # Output path
        subdir = self._engine.build.processing['variables'].get('subdir')
        if subdir:
            name = relpath(name, commonprefix(tuple(
                self._engine.build.pack['files']) + (name,)))
            output = subdir.replace('%(fid)s', camel_case(self.fid))\
                     .replace('%(path)s', dirname(name))
            output = join(self._engine.build.path, 'Output', output)
        else:
            output = join(self._engine.build.path, 'Output')

        # Already processed?
        target = self._engine.config('Output', 'format')
        if target:
            target = join(output, target.format(fid=self.fid))
            if mtime is not None and \
                   not self._engine.build.processing['variables'].get('force')\
                   and exists(target) and getmtime(target) > mtime:
                if not 'files' in self._engine.build.result:
                    self._engine.build.result['files'] = []
                self._engine.build.result['files'].append(relpath(
                    target, join(self._engine.build.path, 'Output')))
                self._engine.build.log(
                    _('${f}: Already processed', {'f': self.fid}))
                return
            if exists(target):
                remove(target)

        # Initialize subdirectory
        if subdir:
            self._engine.initialize(output)

        return output

    # -------------------------------------------------------------------------
    def _script_transform(self, kind, filename):
        """Customized script transformation.

        :param kind: ('preprocess' or 'postprocess')
            Kind of regular expressions.
        :param filename: (string)
            Relative path to file to transform.
        """
        if not kind in self._engine.scripts:
            return
        message = {
            'preprocess': _('${f}: preprocess', {'f': self.fid}),
            'postprocess': _('${f}: post process', {'f': self.fid})}\
            .get(kind, '%s: %s' % (self.fid, kind))
        self._engine.build.log(
            message, 'a_build', self._engine.build.result['log'][-1][2])
        self.data = self._engine.scripts[kind](
            self._engine, filename, self.fid, self.data, self.output)
        if self.data is not None:
            self._save_tmp(1 if kind == 'preprocess' else 5)

    # -------------------------------------------------------------------------
    def _regex_transform(self, kind):
        """Regular expression transformation.

        :param kind: ('preregex' or 'postregex')
            Kind of regular expressions.
        :return: (string)
            Modified data or ``None`` if fails.
        """
        if not kind in self._regex:
            return
        message = {
            'preregex': _('${f}: pre-regex', {'f': self.fid}),
            'postregex': _('${f}: post regex', {'f': self.fid})}\
            .get(kind, '%s: %s' % (self.fid, kind))
        self._engine.build.log(
            message, 'a_build', self._engine.build.result['log'][-1][2])

        # Eventually, convert data into string
        if not isinstance(self.data, basestring):
            method = self._engine.config('Xslt', 'method', 'xml')
            if method in ('text', 'value') or self.data.getroot() is None:
                self.data = str(self.data)
            else:
                self.data = etree.tostring(self.data, method=method,
                    encoding='utf-8', xml_declaration=True, pretty_print=True)

        # Transform
        for regex in self._regex[kind]:
            self.data = regex[0].sub(regex[1], self.data)
        self._save_tmp(2 if kind == 'preregex' else 4)

    # -------------------------------------------------------------------------
    def _xsl_transform(self, variables):
        """XSL transformation.

        :param variables: (dictionary)
            Variable dictionary for XSL.
        """
        if self._xslt is None:
            return
        self._engine.build.log(
            _('${f}: XSL transformation', {'f': self.fid}),
            'a_build', self._engine.build.result['log'][-1][2])

        # Eventually, load XML
        if isinstance(self.data, basestring):
            relaxngs = self._engine.config('Input', 'validate') == 'true' and \
                       self._engine.relaxngs or None
            self.data = load(self.fid, relaxngs, self.data)
            if isinstance(self.data, basestring):
                self._engine.build.stopped(self.data, 'a_error')
                self.data = None
                return

        # Create params dictionary
        params = {}
        for name, value in variables.items():
            if isinstance(value, bool):
                params[name] = str(int(value))
            elif isinstance(value, int):
                params[name] = str(value)
            else:
                params[name] = '"%s"' % value

        # Transform
        # pylint: disable = I0011, W0142
        try:
            self.data = self._xslt(self.data,
                fid='"%s"' % self.fid, output='"%s/"' % self.output, **params)
        except etree.XSLTApplyError, err:
            self._engine.build.stopped(err)
            return

        # Clean xmlns=""
        if self.data.getroot() is not None:
            self.data = etree.tostring(self.data, method='xml',
                encoding='utf-8', pretty_print=True,
                xml_declaration=True).replace(' xmlns=""', '')
        for path, name, files in walk(self.output):
            for name in [join(path, k) for k in files
                         if k[-3:] in ('tml', 'opf', 'ncx')]:
                with open(name, 'r') as hdl:
                    content = hdl.read()
                with open(name, 'w') as hdl:
                    hdl.write(content.replace(' xmlns=""', ''))

        self._save_tmp(3)

    # -------------------------------------------------------------------------
    def _ini_scripts(self, filename):
        """Browse generated INI files and process them.

        :param filename: (string)
            Relative path to the original file to transform.
        """
        for path, name, files in walk(self.output):
            for name in fnmatch.filter(files, '%s-*~.ini' % self.fid):
                self._engine.build.log(_('${f}: script "${n}"',
                    {'f': self.fid, 'n': name[:-5]}),
                    'a_build', self._engine.build.result['log'][-1][2])
                self._iniscript.execute(
                    self.output, filename, join(path, name))
                if self._engine.build.stopped():
                    return

    # -------------------------------------------------------------------------
    def _save_tmp(self, suffix):
        """Save temporary data on file to debug.

        :param suffix: (string)
            Suffix for temporary file name.
        """
        if self._engine.build.processing['variables'].get('keeptmp'):
            fmt = '{fid}.data%d~.%s' % (
                suffix, 'txt' if isinstance(self.data, basestring) else 'xml')
            self._save_data(fmt)

    # -------------------------------------------------------------------------
    def _save_data(self, fmt):
        """Save data on file.

        :param fmt: (string)
            Target name format.
        :return: (string)
            Full path to saved file.
        """
        # Nothing to save
        method = self._engine.config('Xslt', 'method', 'xml')
        if self.data is None or not fmt or method == 'value':
            return
        filename = join(self.output, unicode(fmt).format(fid=self.fid))

        # Save string
        if isinstance(self.data, basestring):
            if not self.data.strip():
                return
            with open(filename, 'w') as hdl:
                hdl.write(self.data)
        # Save XML in text file
        elif method == 'text' or self.data.getroot() is None:
            if self.data.getroot() is not None:
                data = etree.tostring(
                    self.data, encoding='utf-8', pretty_print=True)
            else:
                data = re.sub(r'^<\?xml [^>]*>\s*', '', str(self.data))
            if not data.strip():
                return
            with open(filename, 'w') as hdl:
                hdl.write(data)
        # Save XML/HTML file
        else:
            try:
                self.data.write(filename, method=method, encoding='utf-8',
                                xml_declaration=True, pretty_print=True)
            except (ValueError, AssertionError), err:
                self._engine.build.stopped(err, 'a_error')
                return

        return filename

    # -------------------------------------------------------------------------
    def _finalize(self):
        """Finalization."""
        # Save file
        fmt = self._engine.config('Output', 'format')
        filename = \
            self._save_data(fmt) or join(self.output, fmt.format(fid=self.fid))
        if not exists(filename):
            filename = None

        # Validation
        if filename is not None and self._engine.relaxngs is not None \
               and self._engine.config('Output', 'validate') == 'true':
            self.data = load(filename, self._engine.relaxngs, self.data)
            if isinstance(self.data, basestring):
                self._engine.build.stopped(self.data, 'a_error')
                remove(filename)
                return

        # Container
        container = self._engine.config('Output', 'container') or \
            (self._engine.build.processing['variables'].get('zip') and 'Zip') \
            or None
        if container is not None:
            if container not in self._factories:
                self._engine.build.stopped(
                    _('Unknown container "${c}"', {'c': container}))
                return
            self._engine.build.log(
                _('${f}: container ${c}', {'f': self.fid, 'c': container}),
                'a_build', self._engine.build.result['log'][-1][2])
            filename = self._factories[container].make(self.fid, self.output)
            if not filename:
                return

        # Main finalization
        if self._engine.build.processing['variables'].get('subdir'):
            self._engine.finalize(self.output)

        # Filename in result
        if not self._engine.build.stopped():
            if filename is not None:
                if not 'files' in self._engine.build.result:
                    self._engine.build.result['files'] = []
                self._engine.build.result['files'].append(
                    relpath(filename, join(self._engine.build.path, 'Output')))
            elif isinstance(self.data, basestring):
                if not 'values' in self._engine.build.result:
                    self._engine.build.result['values'] = []
                self._engine.build.result['values'].append(self.data)



# $Id: file.py ada86c3ea608 2012/03/22 22:52:47 patrick $
# pylint: disable = I0011, C0322
"""File view callables."""

from os import sep
from os.path import join, isdir, splitext, relpath
from subprocess import Popen, PIPE
from colander import Mapping, SchemaNode, String, Boolean

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound, HTTPForbidden
from pyramid.response import Response
from pyramid.i18n import get_localizer

from ..lib.utils import _, EXCLUDED_FILES, FILE_TYPES, TEXT_TYPES
from ..lib.utils import get_mime_type, age
from ..lib.views import get_action, current_storage, file_download
from ..lib.views import create_pack, add2container
from ..lib.form import Form
from ..lib.renderer import Paging
from ..models import close_dbsession


# =============================================================================
class FileView(object):
    """Class to manage files in a storage."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='file_search',
                 renderer='../Templates/fil_search.pt',
                 permission='stg.read')
    def search(self):
        """Search file in storages."""
        # Scope
        scope = []
        storage_ids = []
        session = self._request.session
        route = self._request.route_path('storage_root', storage_id='')
        for item in session['menu']:
            for subitem in item[3] or ():
                if subitem[1] and subitem[1].startswith(route):
                    scope.append((subitem[1][len(route):], subitem[0]))
                    storage_ids.append(subitem[1][len(route):])
        if 'storage' in session \
               and not session['storage']['storage_id'] in storage_ids:
            scope.insert(0, (session['storage']['storage_id'],
                              session['storage']['label']))
            storage_ids.append(session['storage']['storage_id'])

        # Form & results
        form = self._search_form(storage_ids)
        results = None
        if form.validate():
            results = self._search(form.values)
        elif 'search' in session:
            results = session['search'][1]

        # Action
        action, items = get_action(self._request)
        if action[0:4] == 'dnl!':
            return file_download(self._request,
                self._request.registry.settings['storage.root'], items,
                get_localizer(self._request).translate(_('search')))
        elif action == 'cls!' and 'container' in session:
            del session['container']
        elif action[0:4] in ('pck!', 'prc!'):
            add2container(self._request, action[0:3], action[4:7], '.', items)
            form.forget('#')
        elif action[0:4] == 'npk!' and 'project' in self._request.session:
            items = create_pack(self._request, items)
            if items[0] is not None:
                return HTTPFound(self._request.route_path(
                    'pack_edit', project_id=items[0], pack_id=items[1]))

        self._request.breadcrumbs.add(_('Advanced search'))
        return {
            'form': form, 'scope': scope, 'FILE_TYPES': FILE_TYPES,
            'results': results}

    # -------------------------------------------------------------------------
    @view_config(route_name='file_info',
                 renderer='../Templates/fil_info.pt',
                 permission='stg.read')
    def info(self):
        """Show file information and VCS log."""
        storage, handler = current_storage(self._request)
        path = self._request.matchdict['path']
        filename = path[-1]
        path = '/'.join(path[:-1])
        fullname = join(handler.vcs.path, path, filename).encode('utf8')
        if self._request.params.get('page_size'):
            self._request.session['page_size'] = \
                int(self._request.params['page_size'])
        log = handler.vcs.log(path, filename,
            int(self._request.session.get('page_size', 20)))

        content = None
        action = get_action(self._request)[0]
        if action[0:4] == 'shw!':
            action = action[4:]
            content = handler.vcs.revision(fullname, action)
            if content is None:
                self._request.session.flash(_('Unable to retrieve this '
                    'revision: it was probably moved.'), 'alert')
            else:
                content = {'route': 'file_revision', 'revision': action,
                    'file': content.decode('utf8'),
                    'label': _('Revision ${r}:', {'r': action})}
        elif action[0:4] == 'dif!':
            action = action[4:]
            content = handler.vcs.diff(fullname, action)
            content = {
                'route': 'file_diff', 'revision': action, 'file': content,
                'label': _('Differences with revision ${r}:', {'r': action})}

        self._request.breadcrumbs.add(_('File information'))
        return {
            'form': Form(self._request), 'storage': storage, 'path': path,
            'filename': filename, 'mime_type': get_mime_type(fullname),
            'TEXT_TYPES': TEXT_TYPES, 'FILE_TYPES': FILE_TYPES, 'log': log,
            'age': age, 'page_sizes': Paging.page_sizes, 'content': content}

    # -------------------------------------------------------------------------
    @view_config(route_name='file_download', permission='stg.read')
    def download(self):
        """Download a file."""
        storage = current_storage(self._request)[0]
        path = self._request.matchdict['path']
        real_path = join(self._request.registry.settings['storage.root'],
                         storage['storage_id'], *path[0:-1])
        if not isdir(real_path):
            raise HTTPNotFound(comment=_('This file does not exist!'))
        return file_download(self._request, real_path, (path[-1],))

    # -------------------------------------------------------------------------
    @view_config(route_name='file_revision', permission='stg.read')
    def revision(self):
        """Retrieve a revision of a file."""
        handler, filename, fullname, revision = self._current_file()
        content = handler.vcs.revision(fullname, revision)

        if content is None:
            self._request.session.flash(_('Unable to retrieve this revision:'
                ' it was probably moved.'), 'alert')
            return HTTPFound(self._request.route_path(
                'file_info', storage_id=handler.uid,
                path='/'.join(self._request.matchdict['path'])))

        download_name = u'%s.r%s%s' % (
            splitext(filename)[0], revision, splitext(filename)[1])
        response = Response(content, content_type=get_mime_type(fullname)[0])
        response.headerlist.append(('Content-Disposition',
            'attachment; filename="%s"' % download_name.encode('utf8')))
        return response

    # -------------------------------------------------------------------------
    @view_config(route_name='file_diff', permission='stg.read')
    def diff(self):
        """Differences between a version and current version."""
        handler, filename, fullname, revision = self._current_file()
        content = handler.vcs.diff(fullname, revision)

        download_name = u'%s.r%s.diff' % (splitext(filename)[0], revision)
        response = Response(content, content_type='text/x-diff')
        response.headerlist.append(('Content-Disposition',
            'attachment; filename="%s"' % download_name.encode('utf8')))
        return response

    # -------------------------------------------------------------------------
    def _current_file(self):
        """Return handler, filename, fullname, and revision."""
        handler = current_storage(self._request)[1]

        # pylint: disable = I0011, W0142
        path = self._request.matchdict['path']
        revision = self._request.matchdict['revision']
        fullname = handler.vcs.full_path(*path)
        if fullname is None or isdir(fullname):
            raise HTTPForbidden()

        return handler, path[-1], fullname, revision

    # -------------------------------------------------------------------------
    def _search_form(self, storage_ids):
        """Return a search form.

        :param storage_ids: (list)
            Available storage IDs.
        :return: (:class:`~.lib.form.Form` instance)
        """
        # Schema
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='files', missing=None))
        schema.add(SchemaNode(String(), name='words', missing=None))
        for sid in storage_ids:
            schema.add(SchemaNode(Boolean(), name='~%s' % sid, missing=False))

        # Defaults
        defaults = {}
        if 'search' in self._request.session:
            defaults = self._request.session['search'][0]
        elif 'storage' in self._request.session:
            defaults = {'~%s' %
                self._request.session['storage']['storage_id']: True}

        return Form(self._request, schema=schema, defaults=defaults)

    # -------------------------------------------------------------------------
    def _search(self, criteria):
        """Search files in storages and return a result.

        :param criteria: (dictionary)
            Search criteria.
        :return: (list)
            A list of tuples such as ``(score, file_type, storage_id,
            path)``.

        This method stores criteria and results in ``session['search']``.
        """
        if not criteria['files'] and not criteria['words']:
            self._request.session.flash(_('Define search criteria!'), 'alert')
            return

        # Search in storages
        found = False
        results = []
        for storage_id in criteria:
            if storage_id[0] == '~' and criteria[storage_id]:
                found = True
                results += self._search_storage(
                    storage_id[1:], criteria['files'], criteria['words'])

        if not found:
            self._request.session.flash(_('Select a storage!'), 'alert')
            return

        self._request.session['search'] = (criteria, results)
        return results

    # -------------------------------------------------------------------------
    def _search_storage(self, storage_id, files, words):
        """Search files in one storage.

        :param storage_id: (string)
            Id of storage.
        :param files: (string)
            Query on files.
        :param words: (string)
            Words to search.
        :return: (list)
            See :meth:`_search`
        """
        root = join(
            self._request.registry.settings['storage.root'], storage_id)

        # Query by file name
        results1 = self._search_storage_by_filename(root, files)
        if results1 is not None and not results1:
            return results1

        # Query by word
        results2 = self._search_storage_by_word(root, words)
        if results1:
            results2 = set(results1) & set(results2)

        # Complete
        results1 = []
        for item in results2:
            results1.append((100, get_mime_type(item)[1],
                             storage_id, relpath(item, root)))
        return results1

    # -------------------------------------------------------------------------
    @classmethod
    def _search_storage_by_filename(cls, root, files):
        """Search files by file name.

        :param root: (string)
            Root path.
        :param files: (string)
            Query on files.
        :return: (list or ``None``)
        """
        if files is None or not files.strip():
            return

        excluded = set(EXCLUDED_FILES)
        results = []
        files = files.split()
        cmd = ['find', root, '-type', 'f', '-name',  files[0]]
        for item in files[1:]:
            cmd += ['-or', '-name', item]

        for item in Popen(cmd, stdout=PIPE).communicate()[0].split('\n'):
            item = item.strip().decode('utf8')
            if item and excluded.isdisjoint(set(item.split(sep))):
                results.append(item)

        return results

    # -------------------------------------------------------------------------
    @classmethod
    def _search_storage_by_word(cls, root, words):
        """Search files by word in it.

        :param root: (string)
            Root path.
        :param words: (string)
            Words to search.
        :return: (list or ``None``)
        """
        if words is None or not words.strip():
            return

        excluded = set(EXCLUDED_FILES)
        results = {}
        words = words.split()
        for item in words:
            cmd = ['find', root, '-type', 'f',
                   '-exec', 'grep', '-l', item, '{}', ';']
            files = Popen(cmd, stdout=PIPE).communicate()[0].split('\n')
            files = [k.strip().decode('utf8') for k in files if k.strip()]
            if not results:
                for item in files:
                    if excluded.isdisjoint(set(item.split(sep))):
                        results[item] = None
            else:
                for item in results.keys():
                    if not item in files:
                        del results[item]
            if not results:
                break

        return results.keys()



# $Id: setup.py ada86c3ea608 2012/03/22 22:52:47 patrick $

import os

from setuptools import setup, find_packages, Extension

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

requires = [
    'pyramid>=1.3',
    'pyramid_beaker>=0.6',
    'pyramid_rpc>=0.3.1',
    'Chameleon>=2.8',
    'SQLAlchemy>=0.7',
    'psycopg2>=2.4',
    'colander>=0.9.6',
    'lxml>=2.3',
    'pycrypto>=2.4',
    'Babel>=0.9.6',
    'Mercurial>=2.1',
    'WebError>=0.10',
    'WebHelpers>=1.3',
    'WebTest>=1.3',
    ]

setup(name='PubliForge',
      version='0.5.1',
      description='Online multimedia publishing system',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        'Programming Language :: Python',
        'Framework :: Pylons',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        ],
      author='Patrick PIERRE',
      author_email='patrick.pierre@prismallia.fr',
      url='www.publiforge.org',
      keywords='web wsgi bfg pylons pyramid publiforge',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='publiforge',
      install_requires = requires,
      ext_modules = [
          Extension('publiforge.lib.rsync._librsync',
                    ['publiforge/lib/rsync/_librsync.c'],
                    libraries=['rsync'])
          ],
      message_extractors = { '.': [
             ('**.py', 'python', None),
             ('**.pt', 'mako', {'input_encoding': 'utf-8'}),
             ]},
      entry_points = """\
      [paste.app_factory]
      main = publiforge:main
      [console_scripts]
      pfpopulate = publiforge.scripts.pfpopulate:main
      pfbackup = publiforge.scripts.pfbackup:main
      pfbuild = publiforge.scripts.pfbuild:main
      """,
      )
